#ifndef sigma_Env_INCLUDED
#define sigma_Env_INCLUDED

#include <ptr/Ptr.hpp>

namespace sigma {

template <typename Key, typename Value>
struct Env;

template <typename Key, typename Value>
int getDepth(const Ptr<Env<Key,Value> >&);

template <typename Key, typename Value>
int maxDepth(const Ptr<Env<Key,Value> >& a, const Ptr<Env<Key,Value> >& b);

template <typename Key, typename Value>
struct Env {

  // thrown by lookup

  struct NotFound {
    NotFound(const Key& k) : key(k) {}
    Key key;
  };

  Env(const Ptr<Env<Key,Value> >& l, 
      const Ptr<Env<Key,Value> >& r, 
      Key k, const Value& v)

    : left(l), 
      right(r), 
      key(k), 
      value(v), 
      depth(maxDepth<Key,Value>(l,r)+1) {}

  Env(Key k, const Value& v) : key(k), value(v), depth(1) {}

  Ptr<Env> left;
  Ptr<Env> right;

  Key key;
  Value value;
  int depth;
};

template <typename Key, typename Value>
int getDepth(const Ptr<Env<Key,Value> >& env) {

  if (!null(env)) return env->depth;
  return 0;
}


template <typename Key, typename Value>
int maxDepth(const Ptr<Env<Key,Value> >& a, const Ptr<Env<Key,Value> >& b) {
  int d_a = getDepth<Key,Value>(a);
  int d_b = getDepth<Key,Value>(b);
  if (d_b<d_a) return d_a;
  return d_b;
}


/*

  Tree node rotation.

      y          x
     / \        / \
    x   c      a   y
   / \            / \
  a   b          b   c

*/


template <typename Key, typename Value>
Ptr<Env<Key,Value> > extend(const Ptr<Env<Key,Value> >& env, 
			    Key key, 
			    const Value& value) {

  if (null(env)) {
    Ptr<Env<Key,Value> > new_env = new Env<Key,Value>(key, value);
    return new_env;
  }

  if (key < env->key) {

    const Ptr<Env<Key,Value> >& y_env = env;

    Ptr<Env<Key,Value> > x_env = extend(y_env->left, key, value);
    const Ptr<Env<Key,Value> >& c_env = env->right;

    int a_depth = getDepth(x_env->left);
    int c_depth = getDepth(y_env->right);

    if (0 < (a_depth-c_depth)) {

      const Ptr<Env<Key,Value> >& a_env = x_env->left;
      const Ptr<Env<Key,Value> >& b_env = x_env->right;
      const Ptr<Env<Key,Value> >& c_env = y_env->right;

      Ptr<Env<Key,Value> > new_y_env 
	= new Env<Key,Value>(b_env, c_env, y_env->key, y_env->value);

      Ptr<Env<Key,Value> > new_x_env 
	= new Env<Key,Value>(a_env, new_y_env, x_env->key, x_env->value);

      return new_x_env;

    } else {

      Ptr<Env<Key,Value> > new_env = new Env<Key,Value>(x_env, 
							c_env, 
							y_env->key, 
							y_env->value);
      return new_env;

    }

  } else if (env->key < key) {

    const Ptr<Env<Key,Value> >& x_env = env;

    Ptr<Env<Key,Value> > y_env = extend(x_env->right, key, value);
    const Ptr<Env<Key,Value> >& a_env = env->left;

    int c_depth = getDepth(y_env->right);
    int a_depth = getDepth(x_env->left);

    if (0 < (c_depth-a_depth)) {

      const Ptr<Env<Key,Value> >& c_env = y_env->right;
      const Ptr<Env<Key,Value> >& b_env = y_env->left;
      const Ptr<Env<Key,Value> >& a_env = x_env->left;

      Ptr<Env<Key,Value> > new_x_env 
	= new Env<Key,Value>(a_env, b_env, x_env->key, x_env->value);

      Ptr<Env<Key,Value> > new_y_env 
	= new Env<Key,Value>(new_x_env, c_env, y_env->key, y_env->value);

      return new_y_env;

    } else {

      Ptr<Env<Key,Value> > new_env = new Env<Key,Value>(a_env, 
							y_env, 
							x_env->key, 
							x_env->value);
      return new_env;
    }

  } else {

    Ptr<Env<Key,Value> > new_env = new Env<Key,Value>(env->left, 
						      env->right, 
						      key, 
						      value);
    return new_env;
  }
}


template <typename Key, typename Value>
Ptr<Env<Key,Value> > merge(const Ptr<Env<Key,Value> >& env1, 
			   const Ptr<Env<Key,Value> >& env2) {

  assert(!null(env1));
  if (null(env2)) return env1;

  Ptr<Env<Key,Value> > env_new = extend(env1, env2->key, env2->value);
  Ptr<Env<Key,Value> > env_left = merge(env_new, env2->left);
  Ptr<Env<Key,Value> > env_right = merge(env_left, env2->right);

  return env_right;
}

template <typename Key, typename Value>
Value lookup(const Ptr<Env<Key,Value> >& env, Key key) {

  if (null(env)) throw Env<Key,Value>::NotFound(key);

  if (key==env->key) return env->value;
  if (key < env->key) return lookup(env->left, key);
  assert(env->key<key);
  return lookup(env->right, key);
}


template <typename Key, typename Value, typename Nothing>
Value exists(const Ptr<Env<Key,Value> >& env, 
	     Key key, 
	     const Nothing& nothing) {


  if (null(env)) return nothing;

  if (key==env->key) return env->value;
  if (key < env->key) return exists(env->left, key, nothing);
  assert(env->key<key);
  return exists(env->right, key, nothing);
}


template <typename Key, typename Value>
int getSize(const Ptr<Env<Key,Value> >& env) {

  if (!null(env)) return 1 + getSize(env->left) + getSize(env->right);
  return 0;

}



} // end namespace sigma

#endif
