#ifndef sigma_lex_Dfa_INCLUDED
#define sigma_lex_Dfa_INCLUDED

/*

  Builds a Deterministic Finite Automata directly from a regular
  expression.  See section 3.9 of "Compilers: Principles, Techniques,
  and Tools" by Aho, Sethi, and Ullman.

*/

#ifdef SIGMA_LEXER_DEBUG
#include <sstream>
#include <fstream>
#endif


#include <assert.h>

#include <map>
#include <vector>

#include <ptr/Ptr.hpp>
#include <lex/Pos.hpp>
#include <lex/Dstate.hpp>


namespace sigma {

template <typename Letter>
struct Letters : std::vector<Letter> {};

// Action really isn't anything more than a plain function
// pointer. This could be typedef'd in a traits class.

template <typename Letter, typename Token, typename Context>
struct Action {

  typedef Token (*FuncPtr)(const Ptr<Context>&, const Letters<Letter>&);

  Action(FuncPtr f) : func_ptr(f) {}
  Action() : func_ptr(0) {} // is this needed?

  Token operator()(const Ptr<Context>& context,
		   const Letters<Letter>& letters) const {
    return (*func_ptr)(context, letters);
  }

  friend inline bool null(const Action<Letter,Token,Context>& action) {
    return !action.func_ptr;
  }

  FuncPtr func_ptr;
};

namespace dfa_detail {

template <typename Letter, typename Token, typename Context>
struct AcceptPositions : std::map<Pos<Letter>, 
				  Action<Letter,Token,Context> > {};

template <typename Letter, typename Token, typename Context>
struct DfaTmp;

template <typename Letter>
struct NodeData {

  bool nullable;
  PosSet<Letter> first;
  PosSet<Letter> last;
};


template <typename Letter, typename Token, typename Context>
struct Node {

  virtual NodeData<Letter> build(DfaTmp<Letter,Token,Context>&) const = 0;
};


template <typename Letter>
struct FollowMap : std::map<Pos<Letter>, PosSet<Letter> > {};

template <typename Letter>
struct EdgeMap : std::map<Letter, Dstate<Letter> > {};

template <typename Letter>
struct Dtrans : std::map<Dstate<Letter>, EdgeMap<Letter> > {};

} // end namespace dfa_detail

template <typename Letter, typename Token, typename Context>
struct Dfa {

  typedef int State;
  typedef std::map<Letter,State> StateMap;
  typedef std::vector<StateMap> StateMaps;
  typedef std::map<State,Action<Letter,Token,Context> > Actions;

#ifdef SIGMA_LEXER_DEBUG
  mutable std::ofstream debug_out;
#endif

  Ptr<Context> context;
  StateMaps state_maps;
  Actions actions;

  Dfa(const dfa_detail::Dtrans<Letter>& dtrans, 
      const dfa_detail::AcceptPositions<Letter,Token,Context>& accept_positions,
      const dfa_detail::PosSets<Letter>& pos_sets) 
    : state_maps(pos_sets.size()) {

#ifdef SIGMA_LEXER_DEBUG
    std::ostringstream debug_file_name;
    debug_file_name << "lexer_debug_" << this << ".txt";
    debug_out.open(debug_file_name.str().c_str());
    if (!debug_out) {
      std::cerr << "Dfa::Dfa: Can't open " 
		<< debug_file_name.str() << std::endl;
    }
#endif

    for (typename dfa_detail::Dtrans<Letter>::const_iterator i = 
	   dtrans.begin();
	 i != dtrans.end();
	 ++i) {

      const dfa_detail::Dstate<Letter>& from_state = (*i).first;
      const dfa_detail::EdgeMap<Letter>& edge_map = (*i).second;

      StateMap& state_map = state_maps[(*from_state).second];

      for (typename dfa_detail::EdgeMap<Letter>::const_iterator j = 
	     edge_map.begin();
	   j != edge_map.end();
	   ++j) {

	Letter letter = (*j).first;
	const dfa_detail::Dstate<Letter>& to_state = (*j).second;

	state_map.insert(std::pair<Letter,int>(letter,(*to_state).second));
      }
    }


    for (typename dfa_detail::PosSets<Letter>::const_iterator k = 
	   pos_sets.begin();
	 k != pos_sets.end();
	 ++k) {

      int state = (*k).second;
      const dfa_detail::PosSet<Letter>& pos_set = (*k).first;

      for (typename dfa_detail::PosSet<Letter>::const_iterator m = 
	     pos_set.begin();
	   m != pos_set.end();
	   ++m) {

	const dfa_detail::Pos<Letter>& pos = *m;

	typename dfa_detail::AcceptPositions<Letter,Token,Context>::const_iterator f 
	  = accept_positions.find(pos);	
	if (f != accept_positions.end()) {
	  // insert fails if action already associated with state
	  const Action<Letter,Token,Context>& action = (*f).second;
	  std::pair<const State,Action<Letter,Token,Context> > p(state,action);
	  actions.insert(p);
	}
      }
    }
  }

  struct BadEndOfInput {};
  struct BadLetter {
    BadLetter(Letter l) : letter(l) {}
    Letter letter;
  };

  template <typename InputIterator>
  Token lex(const Ptr<Context>& context, 
	    InputIterator& in, 
	    InputIterator end = InputIterator()) const {

    static Letters<Letter> letters;
    letters.clear();
    State state = State();
    assert(!state_maps.empty());

    typename StateMap::const_iterator f;

    if (in!=end) {
      f=state_maps[state].find(*in);
      if (f==state_maps[state].end()) throw BadLetter(*in);
      do {

#ifdef SIGMA_LEXER_DEBUG
	if (debug_out) {
	  Letter l = *in;
	  debug_out << "letter: ";
	  if (isprint(l)) {
	    debug_out << "'" << l << "'" << std::endl;
	  } else {
	    debug_out << int(l) << std::endl;
	  }
	  debug_out << "lex state: " << state << std::endl;
	}
#endif

	letters.push_back(*in);
	state = (*f).second;
      } while ((++in!=end) && 
	       ((f=state_maps[state].find(*in))!=state_maps[state].end()));
    }

    typename Actions::const_iterator g = actions.find(state);
    if (g==actions.end()) {
      if (in==end) throw BadEndOfInput();
      throw BadLetter(*in);
    }

    const Action<Letter,Token,Context>& action = (*g).second;

    // Skip null actions, and also skip null tokens (by
    // recursing). This implies that a bool null(Token) function is
    // defined. This allows for actions (like line number counts) that
    // don't get returned to the parser.

    if (null(action)) return lex(context,in,end);
    Token result = (action)(context, letters);
    if (null(result)) return lex(context,in,end);
    return result;
  }
};


namespace dfa_detail {

template <typename Letter>
struct Unmarked : std::vector<Dstate<Letter> > {

  void add(Dstate<Letter> dstate) {
    push_back(dstate);
  }

  Dstate<Letter> get() {
    assert (!empty());
    Dstate<Letter> dstate = back();
    pop_back();
    return dstate;
  }
};


template <typename Letter, typename Token, typename Context>
struct DfaTmp {

  PosSet<Letter> positions;
  FollowMap<Letter> follow_map;

  MakePos<Letter> makePos;

  AcceptPositions<Letter,Token,Context> accept_positions;
};

} // end namespace dfa_detail


template <typename Letter, typename Token, typename Context>
Ptr<Dfa<Letter,Token,Context> >
makeDfa(const Ptr<dfa_detail::Node<Letter,Token,Context> >& grammar) {

  dfa_detail::MakeDstate<Letter> makeDstate;
  dfa_detail::Unmarked<Letter> unmarked;
  dfa_detail::Dtrans<Letter> dtrans;

  dfa_detail::DfaTmp<Letter,Token,Context> tmp;

  dfa_detail::NodeData<Letter> root = grammar->build(tmp);

  std::pair<dfa_detail::Dstate<Letter>, bool> p = makeDstate(root.first);
  assert(p.second);

  unmarked.add(p.first);
  while (!unmarked.empty()) {

    dfa_detail::Dstate<Letter> t = unmarked.get();

    /*
     Build u's for all the existing letters simultaneously. More
     efficient and avoids the need to know the alphabet explicitly.  */

    typedef std::map<Letter,dfa_detail::PosSet<Letter> > UMap;
    UMap u_map;

    for (typename dfa_detail::PosSet<Letter>::const_iterator j = 
	   (*t).first.begin();
	 j != (*t).first.end();
	 ++j) {

      Letter letter = (*j).letter;

      typename dfa_detail::FollowMap<Letter>::const_iterator k 
	= tmp.follow_map.find(*j);
      if (k!=tmp.follow_map.end()) {
	const dfa_detail::PosSet<Letter>& pos_set = (*k).second;
	u_map[letter].insert(pos_set.begin(), pos_set.end());
      }
    }

    for (typename UMap::const_iterator i = u_map.begin();
	 i != u_map.end();
	 ++i) {

      Letter a = (*i).first;
      const dfa_detail::PosSet<Letter>& u = (*i).second;

      std::pair<dfa_detail::Dstate<Letter>,bool> p = makeDstate(u);
      if (p.second) unmarked.add(p.first);
      dtrans[t][a] = p.first;
    }
  }

  return new Dfa<Letter,Token,Context>(dtrans, 
				       tmp.accept_positions,
				       makeDstate.getPosSets());
};


template <typename Letter, typename Token, typename Context>
struct Epsilon : dfa_detail::Node<Letter,Token,Context> {

  Epsilon() {}

  virtual dfa_detail::NodeData<Letter> 
  build(dfa_detail::DfaTmp<Letter,Token,Context>& dfa) const {

    dfa_detail::NodeData<Letter> result;
    result.nullable = true;
    return result;
  }
};


template <typename Letter, typename Token, typename Context>
struct Leaf : dfa_detail::Node<Letter, Token, Context> {

  Leaf(Letter l) : letter(l) {}

  virtual dfa_detail::NodeData<Letter> 
  build(dfa_detail::DfaTmp<Letter,Token,Context>& dfa) const {

    dfa_detail::Pos<Letter> pos = dfa.makePos(letter);

    dfa_detail::NodeData<Letter> result;
    result.nullable = false;
    dfa.positions.insert(pos);
    result.first.insert(pos);
    result.last.insert(pos);

    return result;
  }

  Letter letter;
};


template <typename Letter, typename Token, typename Context>
struct Accept : dfa_detail::Node<Letter, Token, Context> {

  Accept(Action<Letter,Token,Context> a) : action(a) {}

  virtual dfa_detail::NodeData<Letter> 
  build(dfa_detail::DfaTmp<Letter,Token,Context>& dfa) const {

    dfa_detail::Pos<Letter> pos = dfa.makePos(Letter());

    dfa_detail::NodeData<Letter> result;
    result.nullable = false;
    dfa.positions.insert(pos);
    result.first.insert(pos);
    result.last.insert(pos);
    dfa.accept_positions.insert(std::pair<dfa_detail::Pos<Letter>,
				Action<Letter,Token,Context> >(pos,action));

    return result;
  }

  Action<Letter,Token,Context> action;
};


template <typename Letter, typename Token, typename Context>
struct Bar : dfa_detail::Node<Letter, Token, Context> {

  Bar(const Ptr<dfa_detail::Node<Letter,Token,Context> >& l, 
      const Ptr<dfa_detail::Node<Letter,Token,Context> >& r)
    : left(l), right(r) {}

  virtual ~Bar() {} // shut the compiler up

  virtual dfa_detail::NodeData<Letter> 
  build(dfa_detail::DfaTmp<Letter,Token,Context>& dfa) const {

    dfa_detail::NodeData<Letter> result;
    dfa_detail::NodeData<Letter> l = left->build(dfa);
    dfa_detail::NodeData<Letter> r = right->build(dfa);

    result.nullable = l.nullable || r.nullable;

    result.first.insert(l.first.begin(), l.first.end());
    result.first.insert(r.first.begin(), r.first.end());

    result.last.insert(l.last.begin(), l.last.end());
    result.last.insert(r.last.begin(), r.last.end());

    return result;
  }
  Ptr<dfa_detail::Node<Letter,Token,Context> > left;
  Ptr<dfa_detail::Node<Letter,Token,Context> > right;
};


template <typename Letter, typename Token, typename Context>
struct Cat : dfa_detail::Node<Letter,Token,Context> {

  Cat(const Ptr<dfa_detail::Node<Letter,Token,Context> >& l, 
      const Ptr<dfa_detail::Node<Letter,Token,Context> >& r) 
    : left(l), right(r) {}

  virtual ~Cat() {} // shut the compiler up

  virtual dfa_detail::NodeData<Letter> 
  build(dfa_detail::DfaTmp<Letter,Token,Context>& dfa) const {

    dfa_detail::NodeData<Letter> result;
    dfa_detail::NodeData<Letter> l = left->build(dfa);
    dfa_detail::NodeData<Letter> r = right->build(dfa);

    result.nullable = l.nullable && r.nullable;

    if (l.nullable) {
      result.first.insert(l.first.begin(),l.first.end());
      result.first.insert(r.first.begin(), r.first.end());
    } else {
      result.first = l.first;
    }

    if (r.nullable) {
      result.last.insert(r.last.begin(), r.last.end());
      result.last.insert(l.last.begin(), l.last.end());
    } else {
      result.last = r.last;
    }

    for (typename dfa_detail::PosSet<Letter>::const_iterator i = 
	   dfa.positions.begin();
	 i != dfa.positions.end();
	 ++i) {

      typename dfa_detail::PosSet<Letter>::const_iterator j = l.last.find(*i);
      if (j!=l.last.end()) {
	
	dfa.follow_map[*i].insert(r.first.begin(), r.first.end());
      }
    };

    return result;
  }

  Ptr<dfa_detail::Node<Letter,Token,Context> > left;
  Ptr<dfa_detail::Node<Letter,Token,Context> > right;
};


template <typename Letter, typename Token, typename Context>
struct Star : dfa_detail::Node<Letter,Token,Context> {

  Star(const Ptr<dfa_detail::Node<Letter,Token,Context> >& c) : child(c) {}
  virtual ~Star() {} // shut the compiler up

  virtual dfa_detail::NodeData<Letter> 
  build(dfa_detail::DfaTmp<Letter,Token,Context>& dfa) const {

    dfa_detail::NodeData<Letter> result;
    dfa_detail::NodeData<Letter> c = child->build(dfa);

    result.nullable = true;
    result.first = c.first;
    result.last = c.last;

    for (typename dfa_detail::PosSet<Letter>::const_iterator i = 
	   dfa.positions.begin();
	 i != dfa.positions.end();
	 ++i) {

      typename dfa_detail::PosSet<Letter>::const_iterator j = 
	result.last.find(*i);
      if (j!=result.last.end()) {
	dfa.follow_map[*i].insert(result.first.begin(), 
				  result.first.end());
      }
    };

    return result;
  }

  Ptr<dfa_detail::Node<Letter,Token,Context> > child;
};


template <typename Letter, typename Token, typename Context>
Ptr<dfa_detail::Node<Letter,Token,Context> > 
accept(Action<Letter,Token,Context> action) {
  return new Accept<Letter,Token,Context>(action);
}


template <typename Letter, typename Token, typename Context>
Ptr<dfa_detail::Node<Letter,Token,Context> > leaf(Letter letter) {
  return new Leaf<Letter,Token,Context>(letter);
}


template <typename Letter, typename Token, typename Context>
Ptr<dfa_detail::Node<Letter,Token,Context> > epsilon() {
  return new Epsilon<Letter,Token,Context>();
}


template <typename Letter, typename Token, typename Context>
Ptr<dfa_detail::Node<Letter,Token,Context> > 
bar(const Ptr<dfa_detail::Node<Letter,Token,Context> >& left,
    const Ptr<dfa_detail::Node<Letter,Token,Context> >& right) {

  return new Bar<Letter,Token,Context>(left,right);
}


template <typename Letter, typename Token, typename Context>
Ptr<dfa_detail::Node<Letter,Token,Context> > 
operator |(const Ptr<dfa_detail::Node<Letter,Token,Context> >& left,
	   const Ptr<dfa_detail::Node<Letter,Token,Context> >& right) {

  return new Bar<Letter,Token,Context>(left,right);
}


template <typename Letter, typename Token, typename Context>
Ptr<dfa_detail::Node<Letter,Token,Context> > 
cat(const Ptr<dfa_detail::Node<Letter,Token,Context> >& left,
    const Ptr<dfa_detail::Node<Letter,Token,Context> >& right) {

  return new Cat<Letter,Token,Context>(left,right);
}


template <typename Letter, typename Token, typename Context>
Ptr<dfa_detail::Node<Letter,Token,Context> > 
operator &(const Ptr<dfa_detail::Node<Letter,Token,Context> >& left,
	   const Ptr<dfa_detail::Node<Letter,Token,Context> >& right) {

  return new Cat<Letter,Token,Context>(left,right);
}


template <typename Letter, typename Token, typename Context>
Ptr<dfa_detail::Node<Letter,Token,Context> > 
star(const Ptr<dfa_detail::Node<Letter,Token,Context> >& child) {

  return new Star<Letter,Token,Context>(child);
}


template <typename Letter, typename Token, typename Context>
Ptr<dfa_detail::Node<Letter,Token,Context> > 
plus(const Ptr<dfa_detail::Node<Letter,Token,Context> >& child) {

  return cat(child,star(child));
}

template <typename Letter, typename Token, typename Context>
Ptr<dfa_detail::Node<Letter,Token,Context> > 
opt(const Ptr<dfa_detail::Node<Letter,Token,Context> >& child) {

  return bar(child,epsilon<Letter,Token,Context>());
}

} // end namespace sigma

#endif
