#ifndef sigma_Integer_INCLUDED
#define sigma_Integer_INCLUDED

#include <iosfwd>
#include <number/Sign.hpp>
#include <number/Natural.hpp>


namespace sigma {

template <Base base=256>
struct Integer {

  Natural<base> natural;
  Sign sign;

  Integer(const Integer<base>& that)
    : natural(that.natural), sign(natural.isZero()?POS:that.sign) {
  }

  template <Base that_base>
  Integer(const Integer<that_base>& that)
    : natural(that.natural), sign(that.sign) {}

  Integer(const Natural<base>& n) : natural(n), sign(POS) {}
  Integer(const Natural<base>& n, Sign s) 
    : natural(n), sign(natural.isZero()?POS:s) {}
  Integer(const Digits& digits, Sign s=POS) 
    : natural(digits), sign(natural.isZero()?POS:s) {}

  Integer(long long n) : natural((n>=0)?n:-n), sign((n>=0)?POS:NEG) {}

  Integer() : sign(POS) {}
  bool isZero() const { return natural.isZero(); }

#if 0
  long getLong() const {
    long l = natural.getLong();
    return ((sign==POS)?l:-l);
  }
#endif

  operator double() const {

    double d(natural);
    return (sign==POS)?d:-d;
  }

  Integer& operator++() {
    switch (sign) {
    case POS:
      ++natural;
      break;
    case NEG:
      if (natural.isZero()) {
	++natural;
	sign = POS;
      } else {
	--natural;
      }
      break;
    }
    return *this;
  }

  Integer& operator--() {
    switch (sign) {
    case NEG:
      ++natural;
      break;
    case POS:
      if (natural.isZero()) {
	++natural;
	sign = NEG;
      } else {
	--natural;
      }
      break;
    }
    return *this;
  }

};

template <Base base>
Integer<base> operator-(const Integer<base>& that) {
    return Integer<base>(that.natural, -that.sign);
}

template <Base base>
bool operator<(const Integer<base>& a, const Integer<base>& b) {

  // Assumes that the sign is always positive if the natural component
  // is zero.

  if (a.sign==POS) {
    if (b.sign==POS) {
      return a.natural<b.natural;
    } else {
      return false;
    }
  } else {
    if (b.sign==POS) {
      return true;
    } else {
      return b.natural<a.natural;
    }
  }
}

template <Base base>
bool operator==(const Integer<base>& a, const Integer<base>& b) {

  // Assumes that the sign is always positive if the natural component
  // is zero.

  return (a.sign==b.sign) && (a.natural==b.natural);
}


template <Base base>
Integer<base> operator+(const Integer<base>& a, const Integer<base>& b) {

  if (a.sign==POS) {
    if (b.sign==POS) {
      return Integer<base>(a.natural+b.natural, POS);
    } else {
      if (b.natural<a.natural) {
	return Integer<base>(a.natural-b.natural, POS);
      } else {
	return Integer<base>(b.natural-a.natural, NEG);
      }
    }
  } else {
    if (b.sign==NEG) {
      return Integer<base>(a.natural+b.natural, NEG);
    } else {
      if (a.natural<b.natural) {
	return Integer<base>(b.natural-a.natural, POS);
      } else {
	return Integer<base>(a.natural-b.natural, NEG);
      }
    }
  }
}
	
	
template <Base base>
Integer<base> operator-(const Integer<base>& a, const Integer<base>& b) {

  return a + -b;
}


template <Base base>
Integer<base> operator*(const Integer<base>& a, const Integer<base>& b) {

  return Integer<base>(a.natural*b.natural, a.sign*b.sign);
}

// Not sure why this is necessary.
// Shouldn't the cast to Integer be automatic?
template <Base base>
Integer<base> operator*(const Integer<base>& a, const Natural<base>& b) {
  return Integer<base>(a.natural*b, a.sign);
}

// ditto
template <Base base>
Integer<base> operator*(const Natural<base>& a, const Integer<base>& b) {
  return Integer<base>(a*b.natural, b.sign);
}

template <Base base>
Integer<base> operator/(const Integer<base>& a, const Integer<base>& b) {

  return Integer<base>(a.natural/b.natural, a.sign*b.sign);
}

template <Base base>
Integer<base> operator%(const Integer<base>& a, const Integer<base>& b) {

  return Integer<base>(a.natural%b.natural, a.sign*b.sign);
}

template <Base base>
QuotRem<Integer<base>,Integer<base> > divmod(const Integer<base>& a,
					     const Integer<base>& b) {
  QuotRem<Natural<base>,Natural<base> > qr = divmod(a.natural,b.natural);
  Sign sign = a.sign*b.sign;
  Integer<base> quot(qr.quot, sign);
  Integer<base> rem(qr.rem, a.sign);

  // round to negative infinity, not zero

  if (sign==NEG) {
    --quot;
    rem = rem + b;
  }

  return QuotRem<Integer<base>,Integer<base> > (quot,rem);
}


template <Base base>
Integer<base> gcd(const Integer<base>& m, const Integer<base>& n) {

  if (m.isZero()) return n;
  return gcd(n % m, m);
}


std::ostream& operator<<(std::ostream&, const Integer<10>&);
std::ostream& operator<<(std::ostream&, const Integer<256>&);

} // end namespace sigma

#endif
