#ifndef sigma_Natural_INCLUDED
#define sigma_Natural_INCLUDED

#include <assert.h>

#include <iosfwd>

#include <deque>
#include <algorithm>

#include <math.h>

namespace sigma {

typedef unsigned char Digit;
typedef int Base; // should assert valid range
typedef std::deque<Digit> Digits;
static const Digit zero='\000';
static const Digit one='\001';

/* 
   The math functions defined for Natural assume that there are no
   leading zeros. The 'trim' function eliminates them. Digits are
   trimmed when used as a constructor argument for Natural, and every
   result is trimmed, so in theory, this is safe. 
*/


inline void trim(Digits& digits) {
  while((digits.size()!=0) && (digits.back()==zero)) {
    digits.pop_back();
  }
}

// for debugging

inline bool trimmed(const Digits& digits) {
  return ((digits.size()==0) || (digits.back()!=0));
}

template <typename S, typename T>
struct QuotRem {

  QuotRem() {}
  QuotRem(const S& q, const T& r) : quot(q), rem(r) {}

  S quot;
  T rem;
};

struct Negative {};
struct DivByZero {};

template <Base base=256>
struct Natural {

  Natural(const Natural<base>& that) : digits(that.digits) {}

  template <Base that_base>
  Natural(const Natural<that_base>& that) {
    assert(base>1 && base<=256);
    Natural<that_base> source = that;
    while(!source.isZero()) {
      QuotRem<Natural<that_base>,Digit> d = divmod(source,base);
      digits.push_back(d.rem);
      source = d.quot;
    }
    trim(digits);
  }


  Natural(long long that) {
    if (that<0) throw Negative();
    while(that!=0) {

      // Should be using 
      //
      //       #include <cstdlib>
      //       std::lldiv_t d = std::div(that,base);
      //
      // but I can't seem to get that to compile.

      long long quot = that / (long long)(base);
      long long rem = that % (long long)(base);
      digits.push_back(rem);
      that = quot;
    }
    trim(digits);
  }
      

  Natural() {assert(base>1 && base<=256);}
  Natural(const Digits& that_digits) : digits(that_digits) {
    assert(base>1 && base<=256);
    trim(digits);
  }

#if 0
  typedef <typename Iterator>
  Natural(Iterator from, Iterator to) : digits(from, to) {
    assert(base>1 && base<=256);
    trim(digits);
  }
#endif
    

  int size() const {return digits.size();}
  bool isZero() const {return (digits.size()==0);}
  bool isOne() const {return ((digits.size()==1) && (digits[0]==one));}

#if 0
  unsigned long getLong() const {

    // XXX install SIGFPE handler here

    unsigned long result = 0;

    for(Digits::reverse_iterator i = digits.rbegin();
	i != digits.rend();
	++i) {
      result = result * base + *i;
    }

    return result;
  }
#endif

  operator double() const {

    if (base!=256) {
      Natural<256> n(*this);
      return double(n);
    }

    assert(sizeof(unsigned long long)>=sizeof(double));

    int size = std::min(sizeof(unsigned long long), digits.size());
    int rest = int(digits.size()) - size;

    unsigned long long sum = 0;

    for (int i = 0; i<size; ++i) {

      unsigned long long digit = digits[digits.size()-i-1];
      sum = sum * base + digit;
    }

    double significand = sum;
    int exp = rest * 8;
    double result = ldexp(significand, exp);
    return result;
  }


  Natural& operator++() {

    if (digits.size()==0) {
      digits.push_back(one);
    } else {
      Digits::iterator i = digits.begin();
      int carry = one;
      while((i!=digits.end()) && (carry!=0)) {
	int sum = int(*i) + carry;
	div_t d = div(sum, base);
	carry = d.quot;
	*i = d.rem;
      }
      if (carry) {
	assert(i==digits.end());
	digits.push_back(carry);
      }
    }
    return *this;
  }

  Natural& operator--() {

    if (isZero()) throw Negative();

    Digits::iterator i = digits.begin();
    int borrow = -one;
    while((i!=digits.end()) && (borrow!=0)) {
      int sum = int(*i) + borrow;
      div_t d = div(sum, base);
      if (d.rem<0) {
	--d.quot;
	d.rem += base;
      }
      borrow = d.quot;
      *i = d.rem;
    }
    assert(borrow==zero);
    return *this;
  }

  Digits digits;
};


template <Base base>
bool operator<(const Natural<base>& a, const Natural<base>& b) {

  assert(trimmed(a.digits) && trimmed(b.digits));

  if (a.size()!=b.size()) {return a.size()<b.size();}

  Digits::const_reverse_iterator i = a.digits.rbegin();
  Digits::const_reverse_iterator j = b.digits.rbegin();

  while(i!=a.digits.rend()) {

    assert(j!=b.digits.rend());

    if (*i < *j) return true;
    if (*i > *j) return false;
    ++i;
    ++j;
  }

  return false;
}


template <Base base>
bool operator==(const Natural<base>& a, const Natural<base>& b) {

  assert(trimmed(a.digits) && trimmed(b.digits));

  if (a.size()!=b.size()) {return false;}

  Digits::const_reverse_iterator i = a.digits.rbegin();
  Digits::const_reverse_iterator j = b.digits.rbegin();

  while(i!=a.digits.rend()) {

    assert(j!=b.digits.rend());

    if (*i != *j) return false;

    ++i;
    ++j;
  }

  return true;
}


template <Base base>
Natural<base> operator+(const Natural<base>& a, const Natural<base>& b) {

  if (b.size()>a.size()) return operator+(b, a);

  Natural<base> result;

  Digits::const_iterator i = a.digits.begin();
  Digits::const_iterator j = b.digits.begin();
  int carry = 0;

  while(j!=b.digits.end()) {

    assert(i!=a.digits.end());

    int sum = int(*i)+int(*j)+carry;
    div_t d = div(sum, base);
    result.digits.push_back(d.rem);
    carry = d.quot;

    ++j;
    ++i;
  }


  while (i!=a.digits.end()) {

    int sum = int(*i)+carry;
    div_t d = div(sum, base);
    result.digits.push_back(d.rem);
    carry = d.quot;

    ++i;
  }

  if (carry) {
    assert(i==a.digits.end());
    result.digits.push_back(carry);
  }

  trim(result.digits);

  return result;
}


template <Base base>
Natural<base> operator-(const Natural<base>& a, const Natural<base>& b) {

  assert(trimmed(a.digits) && trimmed(b.digits));

  if (a<b) throw Negative();

  Natural<base> result;

  Digits::const_iterator i = a.digits.begin();
  Digits::const_iterator j = b.digits.begin();
  int borrow = 0;

  while(j!=b.digits.end()) {

    assert(i!=a.digits.end());

    int dif = int(*i)-int(*j)+borrow;
    div_t d = div(dif, base);
    if (d.rem<0) {
      --d.quot;
      d.rem += base;
    }
    assert(d.rem>=0);
    result.digits.push_back(d.rem);
    borrow = d.quot;

    ++j;
    ++i;
  }


  while (i!=a.digits.end()) {

    int dif = int(*i)+borrow;
    div_t d = div(dif, base);
    if (d.rem<0) {
      --d.quot;
      d.rem += base;
    }
    assert(d.rem>=0);
    result.digits.push_back(d.rem);
    borrow = d.quot;

    ++i;
  }

  assert(borrow==zero);

  trim(result.digits);

  return result;
}


template <Base base>
Natural<base> operator*(const Natural<base>& a, const Natural<base>& b) {

  Natural<base> product;

  if (a.isZero() || b.isZero()) return product;

  Digits& digits(product.digits);

  int a_size = a.digits.size();
  int b_size = b.digits.size();

  int prod_size = a_size + b_size;

  digits.insert(digits.begin(), prod_size, zero);

  Digits::iterator n = digits.begin();
  Digits::const_iterator j = b.digits.begin();

  do {
    int carry = 0;
    Digits::iterator m = n;
    Digits::const_iterator i = a.digits.begin();

    do {

      int prod = int(*i) * int(*j) + int(*m) + carry;
      div_t d = div(prod, base);
      (*m) = d.rem;
      carry = d.quot;

      ++i;
      ++m;

    } while (i!=a.digits.end());

    (*m) = carry;

    ++j;
    ++n;

  } while (j!=b.digits.end());

  trim(product.digits);
      
  return product;
}


template <Base base>
Natural<base> operator*(const Natural<base>& a, Digit d) {

  Natural<base> b;
  b.digits.push_back(d);
  return a * b;
}


template <Base base>
QuotRem<Natural<base>,Digit> divmod(const Natural<base>& a, Base digit) {

  if (digit==zero) throw DivByZero();
  if (a.isZero()) return QuotRem<Natural<base>,Digit>(Natural<base>(), zero);

  Natural<base> result;

  int rem=0;

  Digits::const_reverse_iterator i = a.digits.rbegin();
  do {

    rem = rem*base + *i;
    div_t d = div(rem, digit);
    result.digits.push_front(d.quot);
    rem = d.rem;

    ++i;

  } while (i != a.digits.rend());

  trim(result.digits);

  return QuotRem<Natural<base>,Digit>(result, rem);
}


template <Base base>
QuotRem<Natural<base>,Natural<base> > divmod(const Natural<base>& a, 
					     const Natural<base>& b) {

  assert(trimmed(a.digits) && trimmed(b.digits));

  if (b.isZero()) {
    throw DivByZero();
  }
  if (a.isZero()) {
    return QuotRem<Natural<base>,Natural<base> >(Natural<base>(), 
						 Natural<base>());
  }

  if (b.digits.size()==1) {
    QuotRem<Natural<base>,Digit> qr = divmod(a, b.digits.front());
    Natural<base> rem;
    if (qr.rem!=zero) rem.digits.push_back(qr.rem);
    return QuotRem<Natural<base>,Natural<base> >(qr.quot, rem);
  }
  if (a<b) return QuotRem<Natural<base>,Natural<base> >(Natural<base>(), a);

  QuotRem<Natural<base>,Natural<base> > result;

  /* normalize */

  assert(b.digits.back()!=zero); /* check for divisor leading zeros */

  int d = base/(int(b.digits.back())+1);
  Natural<base> dn = Natural<base>(d);

  Natural<base> u = a * dn;
  Natural<base> v = b * dn;

  u.digits.push_back(zero);

  int v_size = v.digits.size();
  int u_size = u.digits.size();

  int skip = u_size-v_size;

  Natural<base> rem;
  rem.digits = Digits(u.digits.begin()+skip, u.digits.end());
  Digits::reverse_iterator i = u.digits.rbegin()+v_size;

  do {

    rem.digits.push_front(*i);
    trim(rem.digits);

    unsigned long g_n = rem.digits[v_size]*base + rem.digits[v_size-1];
    unsigned long g_d = v.digits[v_size-1];
    unsigned long guess = g_n/g_d;

    Natural<base> prod;
    while (rem < (prod = v * Natural<base>(guess))) {
      --guess;
    }

    rem = rem - prod;
    result.quot.digits.push_front(guess);

  } while (++i != u.digits.rend());

  QuotRem<Natural<base>,Digit> remainder = divmod(rem,d);

  assert(remainder.rem == 0);

  result.rem = remainder.quot;

  // not sure if this is necessary
  trim(result.quot.digits);
  trim(result.rem.digits);

  return result;
}


template <Base base>
Natural<base> operator/(const Natural<base>& a, const Natural<base>& b) {

  QuotRem<Natural<base>,Natural<base> > result = divmod(a,b);
  return result.quot;
}


template <Base base>
Natural<base> operator%(const Natural<base>& a, const Natural<base>& b) {

  QuotRem<Natural<base>,Natural<base> > result = divmod(a,b);
  return result.rem;
}


template <Base base>
Natural<base> pow(const Natural<base>& a, const Natural<base>& b) {

  Natural<base> d(1);

  Natural<2> bits(b);

  for (Digits::reverse_iterator i = bits.digits.rbegin();
       i != bits.digits.rend();
       ++i) {

    d = d * d;

    if (*i) d = d * a;

  }

  return d;
}



template <Base base>
Natural<base> powmod(const Natural<base>& a, 
		     const Natural<base>& b,
		     const Natural<base>& n) {

  Natural<base> d(1);

  Natural<2> bits(b);

  for (Digits::reverse_iterator i = bits.digits.rbegin();
       i != bits.digits.rend();
       ++i) {

    d = (d * d) % n;

    if (*i) d = (d * a) % n;
  }

  return d;
}


template <Base base>
Natural<base> gcd(const Natural<base>& m, const Natural<base>& n) {

  if (m.isZero()) return n;
  return gcd(n % m, m);
}


std::ostream& operator<<(std::ostream&, const Natural<10>&);
std::ostream& operator<<(std::ostream&, const Natural<256>&);	// converts

} // end namespace sigma

#endif
