#ifndef sigma_parse_Parser_INCLUDED
#define sigma_parse_Parser_INCLUDED

#ifdef SIGMA_PARSER_DEBUG
#include <sstream>
#include <fstream>
#endif

#include <vector>
#include <stack>

#include <lex/Lexer.hpp>
#include <parse/Item.hpp>
#include <parse/Productions.hpp>
#include <parse/Dstate.hpp>
#include <parse/Edge.hpp>
#include <parse/Action.hpp>
#include <parse/ProdVal.hpp>

namespace sigma {

template <typename Context>
struct Parser {

#ifdef SIGMA_PARSER_DEBUG
  mutable std::ofstream debug_out;
#endif


  struct BadSymbol {
    BadSymbol(Symbol s) : symbol(s) {}
    Symbol symbol;
  };

  Parser(const Productions<Context>& prods)
    : end_symbol(prods.end_symbol) {

    typedef typename parse_detail::Item<Context>::ItemSet ItemSet;
    typedef typename parse_detail::MakeDstate<Context>::Dstate State;
    typedef typename parse_detail::MakeDstate<Context>::ItemSets States;
    typedef typename parse_detail::SymbolSet SymbolSet;
    typedef typename parse_detail::SymbolSetMap SymbolSetMap;

    typedef typename std::set<parse_detail::Edge> Edges;
    
#ifdef SIGMA_PARSER_DEBUG
    std::ostringstream debug_file_name;
    debug_file_name << "parser_debug_" << this << ".txt";
    debug_out.open(debug_file_name.str().c_str());
    if (!debug_out) {
      std::cerr << "Parser::Parser: Can't open" 
		<< debug_file_name.str() << std::endl;
    }
#endif

    Edges edges;
    parse_detail::MakeDstate<Context> makeDstate;

    std::pair<State, bool> s0 = makeDstate(prods.start());

    assert(s0.second);
    assert((*s0.first).second==0);

    bool changed;

    do {

      changed = false;

      States states = makeDstate.getItemSets();

      for (State i = states.begin(); 
	   i != states.end();
	   ++i) {

	const ItemSet& item_set = (*i).first;
	int item_int = (*i).second;
	for (typename ItemSet::const_iterator j = item_set.begin();
	     j != item_set.end();
	     ++j) {

	  const parse_detail::Item<Context>& item = *j;
	  Symbol symbol = item.pending();
	  ItemSet new_set = prods.goTo(item_set,symbol);
	  if (!null(symbol) && !new_set.empty()) {

	    // To avoid creating an extra state out of the end symbol
	    // edge, this used to test that the symbol wasn't the end
	    // symbol, but then the accept action wasn't being
	    // created. Could create the accept action by hand and
	    // avoid creating the extra state, but there doesn't seem
	    // to be any harm with the way it works now.

	    std::pair<State,bool> s = makeDstate(new_set);
	    changed = changed || s.second;
	    parse_detail::Edge edge(item_int,symbol,(s.first)->second);
	    std::pair<Edges::iterator,bool> e = edges.insert(edge);
	    changed = changed || e.second;
#ifdef SIGMA_PARSER_DEBUG
	    if (e.second) {
	      debug_out << "Edge(" << edge.from << "," 
			<< edge.symbol << "," << edge.to << ")" 
			<< std::endl;
	    }
#endif
	  }
	}
      }
    } while (changed);

    SymbolSetMap follow = prods.buildFollow();
    RulesMap rules_map;

    States states = makeDstate.getItemSets();
    for (State i = states.begin();
	 i != states.end();
	 ++i) {

      const ItemSet& item_set = (*i).first;
      int state = (*i).second;

#ifdef SIGMA_PARSER_DEBUG
      debug_out << "\n***State " << state << std::endl;
      debug_out << item_set << std::endl;
#endif

      ActionMap& action_map = action_table[state];

      for (typename ItemSet::const_iterator j = item_set.begin();
	   j != item_set.end();
	   ++j) {

	const Item& item = *j;
	const Ptr<Production<Context> >& prod = item.production;

	if (prod->getArity()==item.dot) {

	  Symbol return_name = prod->returnName();

#ifdef SIGMA_PARSER_DEBUG
	  debug_out << " Prod: " << return_name << " <-";
	  unsigned int arg_count = prod->getArity();
	  for(unsigned int prod_arg = 0; prod_arg<arg_count; ++prod_arg) {
	    debug_out << " " << prod->argName(prod_arg);
	  }
	  debug_out << "\n follow set:";
#endif

	  SymbolSetMap::const_iterator it = follow.find(return_name);
	  assert(it != follow.end());
	  const SymbolSet& follow_set = (*it).second;
	  for (SymbolSet::const_iterator k = follow_set.begin();
	       k != follow_set.end();
	       ++k) {

	    Symbol symbol = *k;
	    
	    int rule = rulegen(prod,rules_map);

#ifdef SIGMA_PARSER_DEBUG
	    debug_out << " " << symbol;
#endif

	    addAction(state, action_map, symbol, Reduce(rule));

	  }
#ifdef SIGMA_PARSER_DEBUG
	  debug_out << std::endl;
#endif
	}
      }
    }

    for (Edges::const_iterator i = edges.begin();
	 i != edges.end();
	 ++i) {
      const parse_detail::Edge& edge = *i;
      if (edge.symbol==end_symbol) {
	addAction(edge.from, edge.symbol, Accept());
      } else if (prods.isTerminal(edge.symbol)) {
	addAction(edge.from, edge.symbol, Shift(edge.to));
      } else {
	addAction(edge.from, edge.symbol, GoTo(edge.to));
      }
    }
  }

  typedef parse_detail::Shift Shift;
  typedef parse_detail::Reduce Reduce;
  typedef parse_detail::GoTo GoTo;
  typedef parse_detail::Accept Accept;
  typedef parse_detail::Action Action;

  typedef parse_detail::Item<Context> Item;

  typedef std::map<Symbol,Action> ActionMap;
  typedef std::map<int,ActionMap> ActionTable;

  void addAction(int state, 
		 ActionMap& action_map, 
		 Symbol token, 
		 const Action& action) {

    ActionMap::iterator it = action_map.lower_bound(token);
    if ((it!=action_map.end()) &&
	((*it).first==token)) {
      if ((*it).second!=action) {
	std::cout << "Parse conflict on symbol " << token
		  << " in state " << state << std::endl
		  << "was " << (*it).second << std::endl
		  << "wants to be " << action << std::endl;
      } else {
	// already inserted
      }
    } else {
      action_map.insert(it, std::pair<const Symbol,Action>(token, action));
      // add diagnostics here
    }
  }

  void addAction(int state, Symbol token, const Action& action) {

    ActionMap& action_map = action_table[state];
    addAction(state, action_map, token, action);
  }

  typedef std::vector<Ptr<Production<Context> > > Rules;
  typedef std::map<Ptr<Production<Context> >,int> RulesMap;

  int rulegen(const Ptr<Production<Context> >& item, RulesMap& rules_map) {
    
    typename RulesMap::iterator it = rules_map.lower_bound(item);
    if ((it!=rules_map.end()) && ((*it).first.get()==item.get())) {
      return (*it).second;
    }

    int index = rules.size();
    rules.push_back(item);
    rules_map.insert(it, std::pair<const Ptr<Production<Context> >,int>(item,
									index));
    return index;
  }

  typedef std::stack<int> StateStack;
  typedef std::stack<Cell> ValueStack;

  Symbol end_symbol;
  Rules rules;
  ActionTable action_table;

  template <typename Letter, typename InputIterator>
  Cell parse(Lexer<Letter,ProdVal,Context,InputIterator>& lexer) {

    lexer.context->start();

    StateStack state_stack;
    ValueStack value_stack;

    state_stack.push(0);

#ifdef SIGMA_PARSER_DEBUG
      debug_out << "\n\nstarting to parse...\n" << std::endl;
#endif

    while (true) {

      ProdVal token = lexer.lex();
      Symbol symbol = token.symbol;

#ifdef SIGMA_PARSER_DEBUG
      debug_out << "\nstate: " << state_stack.top() << std::endl;
      debug_out << "symbol: " << symbol << std::endl;
#endif

      bool finished = false;
      do {

	int state = state_stack.top();
      
	ActionTable::const_iterator i = action_table.find(state);
	assert(i!=action_table.end());
	const ActionMap& action_map = (*i).second;

	ActionMap::const_iterator j = action_map.find(symbol);

	if (j==action_map.end()) {

	  throw BadSymbol(symbol);

	} else {

	  Action action = (*j).second;

#ifdef SIGMA_PARSER_DEBUG
	  debug_out << "action: " << action << std::endl;
#endif

	  switch (action.type) {

	  case Action::SHIFT:
	    value_stack.push(token.value);
	    state_stack.push(action.state_or_rule);
	    finished = true;
	    break;

	  case Action::REDUCE:
	    {
	      int rule = action.state_or_rule;
	      Production<Context>& prod = *(rules[rule].get());

#ifdef SIGMA_PARSER_DEBUG
	      debug_out << "reduce production: " 
			<< prod.returnName() << std::endl;
#endif

	      unsigned int arity = prod.getArity();
	      assert(arity<=value_stack.size());
	      Args args;
	      for (unsigned int k=0; k<arity; ++k) {
		args.push_front(value_stack.top());
		state_stack.pop();
		value_stack.pop();
	      }

	      Symbol new_symbol = prod.returnName();
	      Cell new_value = prod.apply(lexer.context, args);
	      int old_state = state_stack.top();

	      ActionTable::const_iterator i = action_table.find(old_state);
	      assert(i!=action_table.end());
	      const ActionMap& action_map = (*i).second;
	      ActionMap::const_iterator j = action_map.find(new_symbol);
	      assert(j!=action_map.end());

	      const Action& action = (*j).second;
	      assert(action.type==Action::GOTO);
	      int new_state = action.state_or_rule;

	      state_stack.push(new_state);
	      value_stack.push(new_value);
	    }	    
	    break;

	  case Action::ACCEPT:
	    assert(value_stack.size()>0);
	    lexer.context->finish();
	    return value_stack.top();
	  case Action::GOTO:
	    assert(false);
	    break;
	  }
	}
      } while (!finished);
    }

    // Never gets here. Shut the compiler up.
    return Cell();
  }
};


} // end namespace sigma


#endif
