#ifndef sigma_ptr_Cell_INCLUDED
#define sigma_ptr_Cell_INCLUDED

#include <assert.h>
#include <set>

#ifdef SHOW_COLLECT
#include <iostream>
#endif

#include <ptr/Package.hpp>
#include <init/Init.hpp>
#include <ptr/Type.hpp>

namespace sigma {

/*

  The collect function returns true if objects were actually
  collected. This is used by the final cleanup function which will
  keep calling collect until nothing is collected. In theory this is
  when there is nothing left to collect. In practice, user code may
  have leaked memory containing cells.

  The reason collect may even need to be called more than once is
  because cells may point at user data structures containing pointers
  to hand-allocated data structures containing cells. An initial call
  to collect may clean up the user data structures, but it won't clean
  up the objects pointed at by the cells contained in the
  dynamically-allocated data structures.

*/


bool collect();

struct BadType {

  BadType(Type e, Type f) 
    : expected(e), found(f) {}

  Type expected;
  Type found;
};

struct Cell;

struct Nothing {};

struct Anything {};

template <typename T>
struct Ptr;

namespace cell_detail {

extern bool auto_collect;
extern unsigned int trigger_count;
extern unsigned int survived;
extern unsigned int collected;


typedef std::set<Package*,PackageLessFunctor> Packages;
extern Packages* packages;
extern bool collecting;

void init(bool);

namespace {Init init_cell(&init);}

template <typename T>
Package* getPackage(T* obj) {

  if (obj==0) return 0;

  TypedPackageLocator<T> package(obj);

  Packages::iterator i = packages->lower_bound(&package);

  if ((i == packages->end()) || PackageLessFunctor()(&package,*i)) {

    if (auto_collect) {
      unsigned int collect_level = 
	(survived>(trigger_count/2))?survived*2:trigger_count;
      if (packages->size()>=collect_level) {
	collect();
	survived = packages->size();

	/* can no longer assume that 'i' is valid */

	Package* new_package = new TypedPackage<T>(obj);
	packages->insert(new_package);
	return new_package;
      }
    }

    Package* new_package = new TypedPackage<T>(obj);
    packages->insert(i, new_package);

    return new_package;
  }

  /* else assume most derived class already submitted */

  ++(*i)->refcount;

  return *i;
}

} // end namespace cell_detail

struct Cell {

  /* constructors */

  template <typename T>
  Cell(T* obj) : package(cell_detail::getPackage(obj)) {insert();}

  Cell(const Cell& that) : package(that.package) {
    insert();
    if (package) ++package->refcount;
  }

  Cell() : package(0) {insert();}

  ~Cell() {
    remove();
    if (!cell_detail::collecting) clearPackage();
  }

  friend inline bool null(const Cell& cell) {
    return !cell.package;
  }

  /* assignment */

  template <typename T>
  Cell& operator=(T* obj) {
    clearPackage();
    package = cell_detail::getPackage(obj);
    return *this;
  }

  Cell& operator=(const Cell& that) {
    clearPackage();
    package = that.package;
    if (package) ++package->refcount;
    return *this;
  }

  /* other stuff */

  inline Type type() const {
    return package?(package->type):Type(typeid(Nothing));
  }


  void clear() {
    clearPackage();
    package=0;
  }

 private:

  friend bool collect();

  static Cell* cells;
  Cell* next;
  Cell* prev;

  inline void insert() {
    prev = 0;
    next = cells;
    cells = this;
    if (next) next->prev = this;
  }

  inline void remove() {
    if (next) next->prev = prev;
    if (prev) {
      prev->next = next;
    } else {
      assert(cells == this);
      cells = next;
    }
  }

 public:

  // Should be private but not worth the trouble right now. I think
  // mark, and Ptr<T> are the only things that should see this.

  cell_detail::Package* package;

  inline void clearPackage() {

    assert(!cell_detail::collecting);

    using cell_detail::Packages;
    using cell_detail::packages;

    if (package && --package->refcount==0) {
      Packages::iterator it = packages->find(package);
      if (it!=packages->end()) {
	packages->erase(it);
	delete package;
      } else {
#ifdef SHOW_COLLECT
	std::cout << "[Broke cycle]" << std::endl;
#endif
      }
    }
  }
};


template <typename T>
inline T& deref(const Cell& cell) {

  if (cell.package==0) throw BadType(typeid(T), typeid(Nothing));
  Type type = cell.package->type;
  if (!(Type(typeid(T))==type)) throw BadType(typeid(T), type);
  return *static_cast<T*>(cell.package->ptr);
}

bool operator<(const Cell&, const Cell&);
bool operator==(const Cell&, const Cell&);

} // end namespace sigma

#endif
