#ifndef sigma_ptr_Package_INCLUDED
#define sigma_ptr_Package_INCLUDED

#include <typeinfo>
#include <assert.h>

#include <ptr/Type.hpp>

namespace sigma {

struct Cell;

namespace cell_detail {

struct Package {

  Package(Type t, void* p) 
    : ptr(p), type(t), refcount(1), mark(current_mark) {
  }
  virtual ~Package();
  
  virtual bool contains(Cell const * const) const = 0;
  virtual bool before(Package const * const) const = 0;

  /*
    The 'before' function is to provide a means of ordering packages
    so that they can be stored in a set. The reason it's tricky is
    because different packages might contain pointers to different
    parts of the same compound class instance. They need to be
    considered equal in the sense that one is neither less than or
    greater than the other. That's why the back of the first is
    compared to the front of the second.
  */

  void* ptr;
  Type type;
  unsigned short refcount;
  bool mark;

  static bool current_mark;
};


template <typename T>
struct TypedPackageBase : Package {

  explicit TypedPackageBase(T* ptr) 
    : Package(Type(typeid(T)), ptr) {}

  virtual ~TypedPackageBase() {
    T* obj = reinterpret_cast<T*>(ptr);
    ptr = 0;
    delete obj;
  }

  virtual bool before(Package const * const package) const {
    assert(package);
    void const * const back = reinterpret_cast<T const * const>(ptr)+1;
    return back <= package->ptr;
  }
};


template <typename T>
struct TypedPackage : TypedPackageBase<T> {

  explicit TypedPackage(T* ptr) : TypedPackageBase<T>(ptr) {}

  virtual bool contains(Cell const * const that_cell) const {
    assert(that_cell);
    void const * const that_ptr = that_cell;
    T const * const front_of_next = reinterpret_cast<T const * const>(ptr)+1;
    void const * const back = front_of_next;
    bool result = ((ptr<=that_ptr) && (that_ptr<back));
    return result;
  }
};


/* 
   The TypedPackageLocator doesn't destroy the object. It's used only
   by the 'getPackage' function to find the package container iterator.
*/


template <typename T>
struct TypedPackageLocator : Package {

  explicit TypedPackageLocator(T* ptr) : Package(typeid(T), ptr) {}

  virtual bool contains(Cell const * const that_cell) const {
    void const * const that_ptr = that_cell;
    void const * const back = reinterpret_cast<T const * const>(ptr)+1;
    return ((ptr<=that_ptr) && (that_ptr<back));
  }

  virtual bool before(Package const * const package) const {
    assert(package);
    void const * const back = reinterpret_cast<T const * const>(ptr)+1;
    return back <= package->ptr;
  }
};


struct PackageLessFunctor {
  bool operator()(const Package* a, const Package* b) const {
    return a->before(b);
  }
};

} // end namespace cell_detail

} // end namespace sigma

#endif
