#ifndef sigma_ptr_Ptr_INCLUDED
#define sigma_ptr_Ptr_INCLUDED

#include <typeinfo>

#include <init/Init.hpp>
#include <dag/Dag.hpp>
#include <ptr/Cell.hpp>
#include <ptr/Type.hpp>

namespace sigma {

template <typename T>
struct Ptr : Cell {

  /* constructors */

  template <typename U>
  Ptr(U* obj) : Cell(obj), ptr(obj) {}

  template <typename U>
  Ptr(const Ptr<U>& that) : Cell(that), ptr(that.ptr) {}

#ifdef DEBUG
  virtual ~Ptr() {ptr=0;}
#endif

 private:

  void convert();

 public:

  Ptr(const Ptr<Anything>& that) : Cell(that), ptr(0) {convert();}
  Ptr(const Cell& that) : Cell(that), ptr(0) {convert();}

  Ptr() : ptr(0) {}

  /* assignment */

  template <typename U>
  Ptr<T>& operator=(U* obj) {
    Cell::operator=(obj);
    ptr = obj;
    return *this;
  }

  template <typename U>
  Ptr<T>& operator=(const Ptr<U>& that) {
    Cell::operator=(that);
    ptr = that.ptr;
    return *this;
  }

  Ptr<T>& operator=(const Ptr<Anything>& cell) {
    Cell::operator=(static_cast<const Cell&>(cell));
    ptr = 0;
    convert();
    return *this;
  }

  Ptr<T>& operator=(const Cell& cell) {
    Cell::operator=(cell);
    ptr = 0;
    convert();
    return *this;
  }

  void clear() {
    ptr=0; 
    Cell::clear();
  }

  /* dereference */

  T* get() const {return ptr;}
  T* operator->() const {return get();}
  T& operator*() const {return *get();}

  /* should be private with friend Ptr<U> */

  T* ptr;
};


struct Ptr<Anything> : Cell {

  /* constructors */

  template <typename T>
  Ptr(T* obj) : Cell(obj) {}

  Ptr(const Cell& cell) : Cell(cell) {}

  Ptr(const Ptr<Anything>& that) : Cell(that) {}

  Ptr() {}

  template <typename T>
  Ptr<Anything>& operator=(T* obj) {
    Cell::operator=(obj);
    return *this;
  }

  Ptr<Anything>& operator=(const Cell& that) {
    Cell::operator=(that);
    return *this;
  }

};


struct Ptr<Nothing> : Cell {

  // can't really do anything with a pointer to nothing

};


namespace cell_detail {

struct CastedPackage {

  CastedPackage(Package* p, void* o) : package(p), obj(o) {}

  Package* package;
  void* obj;
};

typedef CastedPackage (*Converter)(const CastedPackage&);

typedef Dag<const std::type_info*, cell_detail::Converter, TypeInfoCompare> 
TypeDag;
extern TypeDag* type_dag;

void initTypeDag(bool);
namespace {Init type_dag_init(&initTypeDag);}

} // end namespace cell_detail


template <typename T>
void Ptr<T>::convert() {

  if (package!=0) {
    
    Type from_type = package->type;

    if (Type(typeid(T))==from_type) {

      ptr = reinterpret_cast<T*>(package->ptr);

    } else {

      cell_detail::TypeDag::Path path 
	= cell_detail::type_dag->getPath(from_type.type_ptr,&typeid(T));
      if (path.size()==0) {
	package=0;
	ptr=0;
      } else {
	cell_detail::CastedPackage casted_package(package, package->ptr);
	for (cell_detail::TypeDag::Path::iterator i = path.begin();
	     i != path.end();
	     ++i) {
	  
	  casted_package = (*i)(casted_package);
	}
	package = casted_package.package;
	ptr = reinterpret_cast<T*>(casted_package.obj);
      }
    }
  }
}


template <typename From, typename To>
struct DeclareCastConverter {

  DeclareCastConverter() {
    cell_detail::type_dag->insert(&typeid(From),&typeid(To),&converter);
  }

  static cell_detail::CastedPackage
  converter(const cell_detail::CastedPackage& from_package) {

    From* from_obj = static_cast<From*>(from_package.obj);
    cell_detail::Package* to_package 
      = cell_detail::getPackage(new To(*from_obj));
    return cell_detail::CastedPackage(to_package, to_package->ptr);
  }
};


template <typename Derived, typename Base>
struct DeclareSubtypeConverter {

  DeclareSubtypeConverter() {
    cell_detail::type_dag->insert(&typeid(Derived),&typeid(Base),&converter);
  }

  static cell_detail::CastedPackage
  converter(const cell_detail::CastedPackage& from_package) {
    Derived* from_obj = static_cast<Derived*>(from_package.obj);
    Base* to_obj = from_obj;
    return cell_detail::CastedPackage(from_package.package, to_obj);
  }
};


inline bool isa(Type a, Type b) {

  if (a==b) return true;
  if (b==typeid(Anything)) return true;
  if (b==typeid(Nothing)) return false;
  if (!cell_detail::type_dag) return false;
  return cell_detail::type_dag->pathExists(a.type_ptr,b.type_ptr);
}

template <typename T>
bool operator<(const Ptr<T>& a, const Ptr<T>& b) {
  return static_cast<const Cell&>(a) < static_cast<const Cell&>(b);
}

template <typename T>
bool operator==(const Ptr<T>& a, const Ptr<T>& b) {
  return static_cast<const Cell&>(a) == static_cast<const Cell&>(b);
}


} // end namespace sigma

#endif
