#include "arith.hpp"

#include <sigma/env.hpp>

#include <math.h>
#include <ptr/Ptr.hpp>
#include <generic/Primitive.hpp>


namespace sigma {

namespace {

template <typename T>
Ptr<T> addT(const Ptr<T>& a, const Ptr<T>& b) {
  return new T(*a + *b);
}


template <typename T>
Ptr<T> subT(const Ptr<T>& a, const Ptr<T>& b) {
  return new T(*a - *b);
}


template <typename T>
Ptr<T> mulT(const Ptr<T>& a, const Ptr<T>& b) {
  return new T(*a * *b);
}


template <typename T>
Ptr<T> divT(const Ptr<T>& a, const Ptr<T>& b) {
  return new T(*a / *b);
}

template <typename T>
Ptr<T> negT(const Ptr<T>& a) {
  return new T(-(*a));
}

template <typename T>
Ptr<bool> lessT(const Ptr<T>& a, const Ptr<T>& b) {
  return new bool(*a < *b);
}

template <typename T>
Ptr<bool> greaterT(const Ptr<T>& a, const Ptr<T>& b) {
  return new bool(*b < *a);
}

template <typename T>
Ptr<bool> equalT(const Ptr<T>& a, const Ptr<T>& b) {
  return new bool(*a == *b);
}

Ptr<bool> notPrim(const Ptr<bool>& a) {
  return new bool(!(*a));
}


Ptr<Integer<> > doubleFloor(const Ptr<double>& x) {

  double d = ::floor(deref<double>(x));

  int e;
  double s = frexp(d, &e);

  if (s==0) return new Integer<>();
  if (e<0) return new Integer<>(-1);

  size_t size = std::min(sizeof(unsigned long long)*8-1, size_t(e));
  size_t rest = e - size;

  long long l = (long long)(ldexp(s, size));	// in theory fits
  bool neg = l<0;
  return new Integer<>(Natural<>(neg?-l:l) * pow(Natural<>(2),
						 Natural<>(rest)), 
		       neg?NEG:POS);
}


Ptr<Integer<> > rationalFloor(const Ptr<Rational<> >& x) {
  Rational<> r = deref<Rational<> >(x);
  return new Integer<>(r.num.natural / r.den, r.num.sign);
}

Ptr<double> doubleSqrt(const Ptr<double>& z) {

  return new double(::sqrt(deref<double>(z)));
}
Ptr<double> doubleSin(const Ptr<double>& z) {

  return new double(::sin(deref<double>(z)));
}
Ptr<double> doubleCos(const Ptr<double>& z) {

  return new double(::cos(deref<double>(z)));
}
Ptr<double> doubleTan(const Ptr<double>& z) {
  return new double(::tan(deref<double>(z)));
}

Ptr<double> doubleAsin(const Ptr<double>& z) {

  return new double(::asin(deref<double>(z)));
}
Ptr<double> doubleAcos(const Ptr<double>& z) {

  return new double(::acos(deref<double>(z)));
}
Ptr<double> doubleAtan(const Ptr<double>& z) {
  return new double(::atan(deref<double>(z)));
}
Ptr<double> doubleAtan2(const Ptr<double>& y, const Ptr<double>& x) {
  return new double(::atan2(deref<double>(y),deref<double>(x)));
}

} // end namespace


void initArith(bool start) {

  if (start) {

    {
      Ptr<Generic> add = getGeneric("+");

      add->bind(primitive(&addT<double>));
      add->bind(primitive(&addT<Rational<> >));
      add->bind(primitive(&addT<Natural<> >));
      add->bind(primitive(&addT<Integer<> >));
    }

    {
      Ptr<Generic> sub = getGeneric("-");

      sub->bind(primitive(&subT<double>));
      sub->bind(primitive(&subT<Rational<> >));
      sub->bind(primitive(&subT<Integer<> >));

      sub->bind(primitive(&negT<double>));
      sub->bind(primitive(&negT<Rational<> >));
      sub->bind(primitive(&negT<Integer<> >));
    }

    {
      Ptr<Generic> mul = getGeneric("*");
      mul->bind(primitive(&mulT<double>));
      mul->bind(primitive(&mulT<Rational<> >));
      mul->bind(primitive(&mulT<Integer<> >));
      mul->bind(primitive(&mulT<Natural<> >));
    }

    {
      Ptr<Generic> div = getGeneric("/");
      div->bind(primitive(&divT<double>));
      div->bind(primitive(&divT<Rational<> >));
    }

    {
      Ptr<Generic> less = getGeneric("<");
      less->bind(primitive(&lessT<double>));
      less->bind(primitive(&lessT<Rational<> >));
      less->bind(primitive(&lessT<Integer<> >));
      less->bind(primitive(&lessT<Natural<> >));
    }

    {
      Ptr<Generic> greater = getGeneric(">");
      greater->bind(primitive(&greaterT<double>));
      greater->bind(primitive(&greaterT<Rational<> >));
      greater->bind(primitive(&greaterT<Integer<> >));
      greater->bind(primitive(&greaterT<Natural<> >));
    }

    {
      Ptr<Generic> equal = getGeneric("=");
      equal->bind(primitive(&equalT<double>));
      equal->bind(primitive(&equalT<Rational<> >));
      equal->bind(primitive(&equalT<Integer<> >));
      equal->bind(primitive(&equalT<Natural<> >));
    }

    {
      Ptr<Generic> not_generic = getGeneric("not");
      not_generic->bind(primitive(&notPrim));
    }

    {
      Ptr<Generic> floor = getGeneric("floor");
      floor->bind(primitive(&doubleFloor));
      floor->bind(primitive(&rationalFloor));
    }
    {
      getGeneric("sqrt")->bind(primitive(&doubleSqrt));
      getGeneric("sin")->bind(primitive(&doubleSin));
      getGeneric("cos")->bind(primitive(&doubleCos));
      getGeneric("tan")->bind(primitive(&doubleTan));
      getGeneric("asin")->bind(primitive(&doubleAsin));
      getGeneric("acos")->bind(primitive(&doubleAcos));
      Ptr<Generic> atan = getGeneric("atan");
      atan->bind(primitive(&doubleAtan));
      atan->bind(primitive(&doubleAtan2));
    }
  }
}


} // end namespace sigma
