#include "eval.hpp"

#if DEBUG
#include <iostream>
#endif

#include <init/Init.hpp>
#include <symbol/Symbol.hpp>
#include <generic/Primitive.hpp>
#include <sigma/Apply.hpp>
#include <sigma/Bind.hpp>
#include <sigma/Block.hpp>
#include <sigma/Lambda.hpp>
#include <sigma/If.hpp>
#include <sigma/Cons.hpp>
#include <sigma/resolve.hpp>
#include <sigma/env.hpp>

#if DEBUG
#include <sigma/Exp.hpp>
#endif

namespace sigma {

Generic eval("eval");

namespace {

// The eval, replace, and resolve generics have to be initialized
// after the components of the parse tree. That's why the init
// function isn't in the header. (Generics only have object semantics
// after they've been initialized.)


void initEval(bool);
Init init_eval(&initEval);


Ptr<Anything> evalAnything(const Ptr<SigmaEnv>&, 
			   const Ptr<Anything>& value) {

#if DEBUG
  std::cout << "evalAnything(" << value.type().name() << ")" << std::endl;
#endif

  return value;
}


Ptr<Anything> evalSymbol(const Ptr<SigmaEnv>& env, 
			 const Ptr<Symbol>& symbol) {

  return lookup(env,deref<Symbol>(symbol));
}


Ptr<Anything> evalLambda(const Ptr<SigmaEnv>& env,
			 const Ptr<Lambda>& old_lambda) {

  Ptr<Anything> new_lambda = resolve(env, old_lambda);
  return new_lambda;
}


Ptr<Anything> evalApply(const Ptr<SigmaEnv>& env, 
			const Ptr<Apply>& apply_ptr) {

  Apply& apply = deref<Apply>(apply_ptr);

#if DEBUG
  std::cout << "evalApply: " << apply.func.type().name() << std::endl;
#endif

  Ptr<Anything> func_cell = eval(env, apply.func);
  Ptr<Function> func = func_cell;

  if (null(func)) throw BadApply(func_cell);

  Args args;
  for (Args::const_iterator i = apply.args->begin();
       i != apply.args->end();
       ++i) {

    args.push_back(eval(env, *i));
  }

  return func->apply(args);
}


Ptr<Anything> evalBlock(const Ptr<SigmaEnv>& env,
			const Ptr<Block>& block) {
#if DEBUG
  std::cout << "evalBlock" << std::endl;
#endif
  

  if (!null(block->rest)) {
    eval(env,block->statement);	// just for side effect
    return eval(env,block->rest);
  }

  return eval(env,block->statement);
}


  /*

  val f = fn(f,args){body f(f,args)};
  val g = fn(args){f(f,args)};

  val g = fn(f){fn(args){f(f,args)}(fn(f,args){body});

  */

Ptr<Anything> evalBind(const Ptr<SigmaEnv>& env,
		       const Ptr<Bind>& old_bind) {

  // in theory, only called from top level

  Formal formal = old_bind->formal;

  Ptr<Anything> value = eval(env, old_bind->exp);
  Ptr<Bind> new_bind = new Bind(formal, value);
  return new_bind;
}

Ptr<Anything> evalIf(const Ptr<SigmaEnv>& env,
		     const Ptr<If>& if_ptr) {

  Ptr<Anything> cond = eval(env, if_ptr->exp0);
  Ptr<bool> cond_bool = cond;
  if (null(cond_bool)) throw If::NotBool(cond);
  Ptr<Anything> result;
  if (*cond_bool) {
    result = eval(env, if_ptr->exp1);
  } else {
    result = eval(env, if_ptr->exp2);
  }
  return result;
}


Ptr<Anything> evalCons(const Ptr<SigmaEnv>& env,
		       const Ptr<Cons>& p) {
  Ptr<Anything> h = eval(env, hd(p));
  Ptr<Anything> t = eval(env, tl(p));
  return pr(h,t);
}


Ptr<Anything> evalThisEnv(const Ptr<SigmaEnv>& env,
			  const Ptr<ThisEnv>& this_env) {

  return merge(this_env->env, env);
}


void initEval(bool init) {

  if (init) {

    eval.bind(primitive(&evalAnything));
    eval.bind(primitive(&evalSymbol));
    eval.bind(primitive(&evalApply));
    eval.bind(primitive(&evalBlock));
    eval.bind(primitive(&evalBind));
    eval.bind(primitive(&evalLambda));
    eval.bind(primitive(&evalIf));
    eval.bind(primitive(&evalCons));
    eval.bind(primitive(&evalThisEnv));
  }
}

} // end namespace


} // end namespace sigma
