#include "lex.hpp"

#include <sstream>

#include <number/Natural.hpp>

namespace sigma {

namespace {

inline int hexValue(char c) {
  if ((c>='0') && (c<='9')) return int(c) - int('0');
  if ((c>='a') && (c<='f')) return int(c) + 10 - int('a');
  if ((c>='A') && (c<='F')) return int(c) + 10 - int('A');
  return -1; // bad character
}

inline int octValue(char c) {
  if ((c>='0') && (c<='7')) return int(c) - int('0');
  return -1; // bad character
}

inline char hexChar(char c) {
  if (c<=9) return c + '0';
  return c + 'a' - char(10);
}

void hexString(std::string& s, char c) {
  char hi = (0xf0 & c) >> 4;
  char lo = 0x0f & c;
  s.push_back(hexChar(hi));
  s.push_back(hexChar(lo));
}

} // end namespace


// Ignores (undisturbed) ill-formed escape sequences. Assumes two
// hex digits and three octal digits (the first being only zero
// through three).


std::string fromEscSeq(const std::string& in) {

  std::string out;

  for (std::string::const_iterator i = in.begin();
       i != in.end();
       ++i) {

    if (*i == '\\') {
      if (++i == in.end()) return out;
      char c = *i;
      switch(c) {
      case 'n':
	out.push_back('\n');
	break;
      case 't':
	out.push_back('\t');
	break;
      case 'v':
	out.push_back('\v');
	break;
      case 'b':
	out.push_back('\b');
	break;
      case 'r':
	out.push_back('\r');
	break;
      case 'f':
	out.push_back('\f');
	break;
      case 'a':
	out.push_back('\a');
	break;
      case '\\':
	out.push_back('\\');
	break;
      case '?':
	out.push_back('\?');
	break;
      case '\'':
	out.push_back('\'');
	break;
      case '\"':
	out.push_back('\"');
	break;
      case 'x':
	{
	  if (++i == in.end()) {
	    out.push_back('\\');
	    out.push_back('x');
	    return out;
	  }
	  char c_1 = *i;
	  int hex_1 = hexValue(c_1);
	  if (hex_1<0) {
	    out.push_back('\\');
	    out.push_back('x');
	    out.push_back(c_1);
	    break; // skip bad
	  }
	  if (++i == in.end()) {
	    out.push_back('\\');
	    out.push_back('x');
	    out.push_back(c_1);
	    return out;
	  }
	  char c_0 = *i;
	  int hex_0 = hexValue(c_0);
	  if (hex_0<0) {
	    out.push_back('\\');
	    out.push_back('x');
	    out.push_back(c_1);
	    out.push_back(c_0);
	    break; // skip bad
	  }
	  out.push_back(char(hex_1*16+hex_0));
	}
	break;
      case '0':
      case '1':
      case '2':
      case '3':
	{
	  int sum = (c - '0') * 64;
	  if (++i == in.end()) {
	    out.push_back('\\');
	    out.push_back(c);
	    return out;
	  }
	  char c_1 = *i;
	  int oct_1 = octValue(c_1);
	  if (oct_1<0) {
	    out.push_back('\\');
	    out.push_back(c);
	    out.push_back(c_1);
	    break; // skip bad
	  }
	  if (++i == in.end()) {
	    out.push_back('\\');
	    out.push_back(c);
	    out.push_back(c_1);
	    return out;
	  }
	  char c_0 = *i;
	  int oct_0 = octValue(c_0);
	  if (oct_0<0) {
	    out.push_back('\\');
	    out.push_back(c);
	    out.push_back(c_1);
	    out.push_back(c_0);
	    break; // skip bad
	  }
	  out.push_back(char(sum + oct_1*8 + oct_0));
	}
	break;
      default:
	// A backslash is being followed by an invalid escape
	// character. Leave the whole thing in.
	out.push_back('\\');
	out.push_back(c); 
      }
    } else {
      out.push_back(*i);
    }
  }

  return out;
}

      

std::string toEscSeq(const std::string& in) {

  std::string out;

  for (std::string::const_iterator i = in.begin();
       i != in.end();
       ++i) {

    char c = *i;

    if ((c>=0x20) && (c<0x7f) && (c!='"') && (c!='\\')) {

      out.push_back(c);
      
    } else {

      switch(c) {
      case '\n':
	out.append("\\n");
	break;
      case '\t':
	out.append("\\t");
	break;
      case '\v':
	out.append("\\v");
	break;
      case '\b':
	out.append("\\b");
	break;
      case '\r':
	out.append("\\r");
	break;
      case '\f':
	out.append("\\f");
	break;
      case '\a':
	out.append("\\a");
	break;
      case '\\':
	out.append("\\\\");
	break;
      case '"':
	out.append("\\\"");
	break;
      default:
	out.append("\\x");
	hexString(out, c);
	break;
      }
    }
  }

  return out;
}


RegExp token(char c) {
  return leaf<char,ProdVal,Context>(c);
}

RegExp range(char from, char to) {
  if (from>to) return range(to, from);
  if (from==to) return token(from);
  return token(from) | range(from+1,to);
}

RegExp literal(const char* s) {

  if (*s) {
    char letter = *s++;
    return token(letter) & literal(s);
  }
  return epsilon<char,ProdVal,Context>();
}


RegExp oneof(const char* s) {

  if (*s) {
    char letter = *s;
    if (*(s+1)) {
      return token(letter) | oneof(s+1);
    } else {
      return token(letter);
    }
  }
  // not supposed to happen
  return epsilon<char,ProdVal,Context>();
}


RegExp rule(RegExp regexp, Action<char,ProdVal,Context> action) {
  return cat<char,ProdVal>(regexp, accept<char,ProdVal,Context>(action));
}


ProdVal la_factor_op(Ptr<Context> const &, Letters<char> const & letters) {
  return ProdVal("factor_op",
		 new Symbol(std::string(letters.begin(), 
					letters.end())));
}


ProdVal la_comp_op(Ptr<Context> const &, Letters<char> const & letters) {
  return ProdVal("comp_op",
		 new Symbol(std::string(letters.begin(), 
					letters.end())));
}


ProdVal la_and_op(Ptr<Context> const &, Letters<char> const & letters) {
  return ProdVal("&");
}

ProdVal la_or_op(Ptr<Context> const &, Letters<char> const & letters) {
  return ProdVal("|");
}


ProdVal la_pos_op(Ptr<Context> const &, Letters<char> const & letters) {
  return ProdVal("+",
		 new Symbol(std::string(letters.begin(), 
					letters.end())));
}

ProdVal la_neg_op(Ptr<Context> const &, Letters<char> const & letters) {
  return ProdVal("-",
		 new Symbol(std::string(letters.begin(), 
					letters.end())));
}

ProdVal la_left_paren(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("(");
}

ProdVal la_right_paren(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal(")");
}

ProdVal la_left_square(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("[");
}

ProdVal la_right_square(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("]");
}

ProdVal la_left_curly(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("{");
}

ProdVal la_right_curly(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("}");
}

ProdVal la_semi(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal(";");
}

ProdVal la_colon(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal(":");
}

ProdVal la_comma(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal(",");
}

ProdVal la_tick(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("'");
}

ProdVal la_let_lit(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("let");
}

ProdVal la_fun_lit(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("fun");
}

ProdVal la_be_lit(Ptr<Context> const &, Letters<char> const & letters) {
  return ProdVal("be");
}

ProdVal la_fn_lit(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("fn");
}

ProdVal la_if_lit(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("if");
}

ProdVal la_then_lit(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("then");
}

ProdVal la_else_lit(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("else");
}

ProdVal la_this_lit(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("this");
}

ProdVal la_end_of_input(Ptr<Context> const &, Letters<char> const &) {
  return ProdVal("end_of_input");
}

ProdVal la_op_ident(Ptr<Context> const &, Letters<char> const & letters) {
  Letters<char>::const_iterator b = letters.begin();
  Letters<char>::const_iterator e = letters.end();
  assert(b!=e);
  ++b; // skip escape
  return ProdVal("ident", new Symbol(std::string(b, e)));
}


ProdVal la_ident(Ptr<Context> const &, Letters<char> const & letters) {
  
  return ProdVal("ident",
		 new Symbol(std::string(letters.begin(), 
					letters.end())));
}


ProdVal la_string(Ptr<Context> const &, Letters<char> const & letters) {

  int size = letters.size();

  std::string translated = fromEscSeq(std::string(&letters[1],
						  &letters[size-1]));
  return ProdVal("string", new std::string(translated));
}


ProdVal la_natural(Ptr<Context> const &, Letters<char> const & letters) {
  
  Digits digits;
  for (Letters<char>::const_iterator i = letters.begin();
       i != letters.end();
       ++i) {

    digits.push_front((*i) - '0');
  }
  Natural<10> n(digits);
  return ProdVal("natural", 
		 new Natural<>(n));
}


ProdVal la_real(Ptr<Context> const &, Letters<char> const & letters) {

  std::string s(&letters[0], letters.size());
  double d;
  std::istringstream in(s);
  in >> d;
  assert(in);

  return ProdVal("real", new double(d));
}


RegExp buildRegExp() {

  const Action<char,ProdVal,Context> no_action;

  RegExp rules;

  RegExp nothing = rule(epsilon<char,ProdVal,Context>(), &la_end_of_input);

  RegExp wsc = oneof(" \n\t\r");
  RegExp ws = rule(plus(wsc), no_action);

  RegExp digit = range('0', '9');
  RegExp digits = plus(digit);
  RegExp lower = range('a', 'z');
  RegExp upper = range('A', 'Z');
  RegExp letter = lower | upper | token('_');

  // Escaped characters. See: http://www-ccs.ucsd.edu/c/charset.html
  // This implementation assumes an escaped octal number is always
  // three digits, and an escaped hex number is always two digits.
  // (You can't just ignore double quotes preceeded by a backslash
  // because it might be an escaped backslash.)

  RegExp hex_digit = digit | oneof("abcdefABCDEF");
  RegExp oct_digit = range('0', '7');
  RegExp escaped_char = token('\\') & oneof("ntvbrfa\\?'\"");
  RegExp escaped_oct = token('\\') & oct_digit & oct_digit & oct_digit;
  RegExp escaped_hex = literal("\\x") & hex_digit & hex_digit;

  RegExp not_double_quote = 
    range(0x20,'"'-1) | 
    range('"'+1,'\\'-1) |
    range('\\'+1,0x7e) |
    escaped_char |
    escaped_oct |
    escaped_hex;

  RegExp string = rule(token('"') & star(not_double_quote) & token('"'),
		       &la_string);

  RegExp op_ident = rule(token('\\') & oneof("+-*/<>="), 
			  &la_op_ident);

  RegExp ident = rule(letter & star(letter | digit),
		       &la_ident);

  RegExp natural = rule(plus(digit), 
			&la_natural);

  RegExp mant = token('.') & digits;
  RegExp exp = oneof("eE") & opt(oneof("+-")) & digits;

  RegExp real = rule(((digits & opt(mant)) | mant) & 
		     opt(exp),
		     &la_real);

  RegExp left_paren = rule(token('('),&la_left_paren);
  RegExp right_paren = rule(token(')'),&la_right_paren);

  RegExp left_square = rule(token('['),&la_left_square);
  RegExp right_square = rule(token(']'),&la_right_square);

  RegExp left_curly = rule(token('{'),&la_left_curly);
  RegExp right_curly = rule(token('}'),&la_right_curly);

  RegExp semi = rule(token(';'),&la_semi);
  RegExp colon = rule(token(':'),&la_colon);

  RegExp comma = rule(token(','),&la_comma);

  RegExp tick = rule(token('\''),&la_tick);

  RegExp factor_op = rule(oneof("*/"),&la_factor_op);

  RegExp comp_op = rule(oneof("<>="), &la_comp_op);

  RegExp and_op = rule(token('&'),&la_and_op);
  RegExp or_op = rule(token('|'),&la_or_op);

  RegExp pos_op = rule(token('+'),&la_pos_op);
  RegExp neg_op = rule(token('-'),&la_neg_op);

  RegExp be_lit = rule(literal("be"),&la_be_lit);

  RegExp let_lit = rule(literal("let"),&la_let_lit);

  RegExp fun_lit = rule(literal("fun"),&la_fun_lit);

  RegExp fn_lit = rule(literal("fn"),&la_fn_lit);

  RegExp if_lit = rule(literal("if"),&la_if_lit);
  RegExp then_lit = rule(literal("then"),&la_then_lit);
  RegExp else_lit = rule(literal("else"),&la_else_lit);
  RegExp this_lit = rule(literal("this"),&la_this_lit);


  rules = 
    nothing 		|
    ws			|
    let_lit		|
    fun_lit		|
    be_lit		|
    fn_lit		|
    if_lit		|
    then_lit		|
    else_lit		|
    this_lit		|
    left_paren		|
    right_paren		|
    left_square		|
    right_square	|
    left_curly		|
    right_curly		|
    semi		|
    colon		|
    comma		|
    tick		|
    factor_op		|
    and_op		|
    or_op		|
    comp_op		|
    pos_op		|
    neg_op		|
    string		|
    natural		|
    real		|
    op_ident		|
    ident;

  return rules;
}

} // end namespace sigma
