#include "parse.hpp"

#include <ptr/Ptr.hpp>

#include "Apply.hpp"
#include "Bind.hpp"
#include "Block.hpp"
#include "Lambda.hpp"
#include "If.hpp"
#include "Cons.hpp"
#include "Exp.hpp"

namespace sigma {

Cell pa_identity(const Ptr<Context>&,
		 const Cell& x) {
  return x;
}

Cell pa_bind(const Ptr<Context>&,
	     const Cell& /* let */, 
	     const Cell& symbol, 
	     const Cell& /* be */, 
	     const Cell& exp) {
  return new Bind(Formal(deref<Symbol>(symbol)), exp);
}

Cell pa_funbind(const Ptr<Context>&,
		const Cell& /* let */, 
		const Cell& symbol, 
		const Cell& formals, 
		const Cell& /* be */, 
		const Cell& body) {

  Ptr<Anything> lambda = new Lambda(formals, body, deref<Symbol>(symbol));
  return new Bind(Formal(deref<Symbol>(symbol)), lambda);
}

Cell pa_if(const Ptr<Context>&,
	   const Cell& /* if */, 
	   const Cell& exp0, 
	   const Cell& /* then */, 
	   const Cell& exp1,
	   const Cell& /* else */, 
	   const Cell& exp2) {
  return new If(exp0, exp1, exp2);
}

Cell pa_apply(const Ptr<Context>&,
	      const Cell& func, 
	      const Cell& /* ( */, 
	      const Cell& args, 
	      const Cell& /* ) */) {
  return new Apply(func, args);
}

Cell pa_apply_0(const Ptr<Context>&,
		const Cell& f, 
		const Cell& /* ( */, 
		const Cell& /* ) */) {
  Apply* apply = new Apply();
  apply->func = f;
  return apply;
}

Cell pa_args(const Ptr<Context>&,
	     const Cell& args, 
	     const Cell& /* , */, 
	     const Cell& exp) {
  deref<Args>(args).push_back(exp);
  return args;
}


Cell pa_arg(const Ptr<Context>&,
	    const Cell& exp) {
  Args* args = new Args();
  args->push_back(exp);
  return args;
}


Cell pa_fn(const Ptr<Context>&,
	   const Cell& /* fn */, 
	   const Cell& formals, 
	   const Cell& block) {
  return new Lambda(formals, block);
}


Cell pa_proto(const Ptr<Context>&,
	      const Cell& /* ( */, 
	      const Cell& formals, 
	      const Cell& /* ) */) {
  return formals;
}


Cell pa_proto_0(const Ptr<Context>&,
		const Cell& /* ( */, 
		const Cell& /* ) */) {
  return new Formals();
}


Cell pa_exp(const Ptr<Context>&,
	    const Cell& /* ' */, 
	    const Cell& exp, 
	    const Cell& /* ' */) {
  return new Exp(exp);
}


Cell pa_formals(const Ptr<Context>&,
		const Cell& formals, 
		const Cell& /* , */, 
		const Cell& formal) {
  deref<Formals>(formals).push_back(deref<Formal>(formal));
  return formals;
}


Cell pa_formals_1(const Ptr<Context>&,
		  const Cell& formal) {
  Formals* formals = new Formals();
  formals->push_back(deref<Formal>(formal));
  return formals;
}


Cell pa_typed_formal(const Ptr<Context>&,
		     const Cell& symbol, 
		     const Cell& /* : */, 
		     const Cell& type) {

  const std::string& name = (deref<Symbol>(type)).name();

  return new Formal(deref<Symbol>(symbol), Type::getTypeByName(name));
}

Cell pa_formal(const Ptr<Context>&,
	       const Cell& symbol) {

  return new Formal(deref<Symbol>(symbol), 
		    Type(typeid(Anything)));
}

Cell pa_andop(const Ptr<Context>&,
	      const Cell& a1, 
	      const Cell& /* & */, 
	      const Cell& a2) {

  Ptr<bool> f = new bool(false);
  Ptr<If> result = new If(a1, a2, f);
  return result;
}

Cell pa_orop(const Ptr<Context>&,
	     const Cell& a1, 
	     const Cell& /* | */, 
	     const Cell& a2) {

  Ptr<bool> t = new bool(true);
  Ptr<If> result = new If(a1, t, a2);
  return result;
}

Cell pa_binop(const Ptr<Context>&,
	      const Cell& a1, 
	      const Cell& f, 
	      const Cell& a2) {
  Apply* apply = new Apply();
  apply->func = f;
  apply->args->push_back(a1);
  apply->args->push_back(a2);
  return apply;
}

Cell pa_unop(const Ptr<Context>&,
	     const Cell& f, 
	     const Cell& a1) {
  Apply* apply = new Apply();
  apply->func = f;
  apply->args->push_back(a1);
  return apply;
}

Cell pa_this_op(const Ptr<Context>&,
		const Cell& t) {
  return new ThisEnv();
}


Cell pa_paren(const Ptr<Context>&,
	      const Cell& /* ( */, 
	      const Cell& x, 
	      const Cell& /* ) */) {
  return x;
}

Cell pa_list(const Ptr<Context>&,
	     const Cell& /* [ */, 
	     const Cell& l, 
	     const Cell& /* ] */) {
  return l;
}

Cell pa_block(const Ptr<Context>&,
	      const Cell& /* { */, 
	      const Cell& block, 
	      const Cell& /* } */) {
  return block;
}


Cell pa_items(const Ptr<Context>&,
	      const Cell& item, 
	      const Cell& /* , */, 
	      const Cell& items) {
  return pr(item,items);
}

Cell pa_item(const Ptr<Context>&,
	     const Cell& item) {
  return pr(item,Ptr<Nil>(new Nil()));
}

Cell pa_nil(const Ptr<Context>&,
	    const Cell& /* [ */, 
	    const Cell& /* ] */) {
  Cell result = Ptr<Nil>(new Nil());
  return result;
}


Cell pa_statement(const Ptr<Context>&,
		  const Cell& statement) {
  return new Block(statement);
}

Cell pa_statements(const Ptr<Context>&,
		   const Cell& stmt, 
		   const Cell& /* , */, 
		   const Cell& stmts) {
  return new Block(stmt, stmts);
}

Productions<Context> buildProductions() {

  Productions<Context> prods("statement", ";");

  prods.add("statement", &pa_identity,	"binding");

  prods.add("binding", &pa_bind,	"let", "ident", "be", "exp");

  prods.add("binding", &pa_identity,	"funbind");

  prods.add("funbind", &pa_funbind,	"let", "ident", "proto", "be", "exp");

  prods.add("statement", &pa_identity,	"exp");

  prods.add("exp", &pa_identity, 	"if_exp");

  prods.add("if_exp", &pa_if,	"if", "exp", "then", "exp", "else", "exp");

  prods.add("if_exp", &pa_identity, 	"bool_exp");

  prods.add("bool_exp", &pa_andop,	"bool_exp", "&", "comp_exp");

  prods.add("bool_exp", &pa_orop,	"bool_exp", "|", "comp_exp");

  prods.add("bool_exp", &pa_identity,	"comp_exp");

  prods.add("comp_exp", &pa_binop,	"term_exp", "comp_op", "term_exp");

  prods.add("comp_exp", &pa_identity,	"term_exp");

  prods.add("term_exp", &pa_binop,	"term_exp", "+", "factor_exp");

  prods.add("term_exp", &pa_binop,	"term_exp", "-", "factor_exp");

  prods.add("term_exp", &pa_identity,	"factor_exp");

  prods.add("factor_exp", &pa_binop,	"factor_exp", "factor_op", "factor");

  prods.add("factor_exp", &pa_identity,	"factor");

  prods.add("factor", &pa_identity,	"string"); 

  prods.add("factor", &pa_identity,	"natural"); 

  prods.add("factor", &pa_identity, 	"real"); 

  prods.add("factor", &pa_identity,	"ident"); 

  prods.add("factor", &pa_this_op,	"this");

  prods.add("factor", &pa_unop,		"-", "natural");

  prods.add("factor", &pa_unop,		"-", "ident");

  prods.add("factor", &pa_paren,        "(", "exp", ")");

  prods.add("factor", &pa_list,		"[", "items", "]");

  prods.add("items", &pa_items,		"exp", ",", "items");

  prods.add("items", &pa_item,		"exp");

  prods.add("factor", &pa_nil,		"[", "]");

  prods.add("factor", &pa_identity,	"block");

  prods.add("factor", &pa_apply,	"factor", "(", "args", ")");

  prods.add("factor", &pa_apply_0,	"factor", "(", ")");

  prods.add("args", &pa_args,		"args", ",", "exp");

  prods.add("args", &pa_arg,		"exp");

  prods.add("factor", &pa_fn,		"fn", "proto", "block");

  prods.add("proto", &pa_proto,		"(", "formals", ")");

  prods.add("proto", &pa_proto_0,	"(", ")");

  prods.add("factor", &pa_exp,		"'", "exp", "'");

  prods.add("formals", &pa_formals,	"formals", ",", "formal");

  prods.add("formals", &pa_formals_1,	"formal");

  prods.add("formal", &pa_typed_formal,	"ident", ":", "ident");

  prods.add("formal", &pa_formal,	"ident");

  prods.add("block", &pa_block,		"{", "statements", "}");

  prods.add("statements", &pa_statements,	"statement", ",", "statements");
  prods.add("statements", &pa_statement,	"statement");

  return prods;
}

} // namespace sigma

