#include "resolve.hpp"

#include <ptr/Ptr.hpp>
#include <init/Init.hpp>

#include <symbol/Symbol.hpp>
#include <generic/Primitive.hpp>
#include <env/Env.hpp>

#include <sigma/env.hpp>
#include <sigma/Block.hpp>
#include <sigma/Cons.hpp>
#include <sigma/Bind.hpp>
#include <sigma/Lambda.hpp>
#include <sigma/Apply.hpp>
#include <sigma/If.hpp>
#include <sigma/replace.hpp>


namespace sigma {

Generic resolve("resolve");

namespace {

// The eval, replace, and resolve generics have to be initialized
// after the components of the parse tree. That's why the init
// function isn't in the header. (Generics only have object semantics
// after they've been initialized.)


void initResolve(bool);
Init init_resolve(&initResolve);

Ptr<Anything> resolveAnything(const Ptr<SigmaEnv>&,
			      const Ptr<Anything>& anything) {

  return anything;
}


Ptr<Anything> resolveSymbol(const Ptr<SigmaEnv>& env, 
			    const Ptr<Symbol>& symbol) {

  Ptr<Anything> value = lookup(env,deref<Symbol>(symbol));

  if (isa(value.type(),Type(typeid(Free)))) {
    return symbol;
  }
  return value;
}


Ptr<Anything> resolveArgs(const Ptr<SigmaEnv>& env,
			  const Ptr<Args>& old_args) {

  Ptr<Args> new_args = new Args();

  for (Args::iterator i = old_args->begin();
       i != old_args->end();
       ++i) {

    new_args->push_back(resolve(env, *i));
  }

  return new_args;
}


Ptr<Anything> resolveApply(const Ptr<SigmaEnv>& env,
			   const Ptr<Apply>& old_apply) {

  Ptr<Anything> func = resolve(env, old_apply->func);
  Ptr<Args> args = resolve(env, old_apply->args);
  Ptr<Apply> new_apply = new Apply(func,args);

  return new_apply;
}


Ptr<Anything> resolveLambda(const Ptr<SigmaEnv>& old_env,
			    const Ptr<Lambda>& old_lambda) {

  if (old_lambda->resolving) return old_lambda;

  old_lambda->resolving = true;

  const Ptr<Formals>& formals = old_lambda->formals;

  Ptr<SigmaEnv> new_env = old_env;

  for (Formals::const_iterator i = formals->begin();
       i != formals->end();
       ++i) {

    new_env = extend(new_env, (*i).symbol, Ptr<Anything>(new Free()));
  }

  Symbol recurse = old_lambda->recurse;

  if (!null(recurse)) {

    Ptr<SigmaEnv> free_env = extend(new_env, 
				    recurse, 
				    Ptr<Anything>(new Free()));
    Ptr<Anything> new_body = resolve(free_env, old_lambda->body);
    old_lambda->resolving = false;
    Ptr<Anything> free_lambda(new Lambda(formals, new_body));
    Ptr<RecLambda> rec_lambda_ptr = new RecLambda(recurse, free_lambda);
    Ptr<Anything> bound_lambda = replace(rec_lambda_ptr, free_lambda);
    return bound_lambda;
  }

  Ptr<Anything> new_body = resolve(new_env, old_lambda->body);

  old_lambda->resolving = false;

  Ptr<Lambda> new_lambda(new Lambda(formals, new_body));

  return new_lambda;
}


Ptr<Anything> resolveBlock(const Ptr<SigmaEnv>& env,
			   const Ptr<Block>& old_block) {

  Ptr<Bind> bind = old_block->statement;

  if (!null(bind)) {

    Ptr<Formals> formals = new Formals();
    formals->push_back(bind->formal);

    Ptr<Lambda> lambda = new Lambda(formals, old_block->rest);

    Ptr<Args> args = new Args();
    args->push_back(bind->exp);

    Ptr<Apply> apply = new Apply(lambda, args);

    Ptr<Anything> result = resolve(env, apply);
    return result;
  }

  Ptr<Anything> statement = resolve(env, old_block->statement);
  Ptr<Anything> rest = resolve(env, old_block->rest);

  Ptr<Block> new_block = new Block(statement,rest);

  return new_block;
}

Ptr<Anything> resolveCons(const Ptr<SigmaEnv>& env,
			  const Ptr<Cons>& old_cons) {

  Ptr<Anything> h = resolve(env,hd(old_cons));
  Ptr<Anything> t = resolve(env,tl(old_cons));

  Ptr<Anything> p = pr(h,t);

  return p;
}


Ptr<Anything> resolveIf(const Ptr<SigmaEnv>& env,
			const Ptr<If>& old_if) {

  Ptr<Anything> exp0 = resolve(env, old_if->exp0);
  Ptr<Anything> exp1 = resolve(env, old_if->exp1);
  Ptr<Anything> exp2 = resolve(env, old_if->exp2);

  Ptr<If> new_if = new If(exp0, exp1, exp2);

  return new_if;
}


Ptr<Anything> resolveBind(const Ptr<SigmaEnv>& env,
			  const Ptr<Bind>& old_bind) {

  // in theory, only called from top level

  Formal formal = old_bind->formal;

  Ptr<Anything> new_exp = resolve(env, old_bind->exp);
  Ptr<Bind> new_bind = new Bind(formal, new_exp);
  return new_bind;
}


Ptr<Anything> resolveThisEnv(const Ptr<SigmaEnv>& env,
			     const Ptr<ThisEnv>& this_env) {


  if (null(this_env->env)) return new ThisEnv(env);

  Ptr<SigmaEnv> new_env = merge(this_env->env, env);
  return new ThisEnv(new_env);
}


void initResolve(bool init) {

  if (init) {
    resolve.bind(primitive(&resolveAnything));
    resolve.bind(primitive(&resolveSymbol));
    resolve.bind(primitive(&resolveArgs));
    resolve.bind(primitive(&resolveApply));
    resolve.bind(primitive(&resolveLambda));
    resolve.bind(primitive(&resolveBlock));
    resolve.bind(primitive(&resolveCons));
    resolve.bind(primitive(&resolveBind));
    resolve.bind(primitive(&resolveIf));
    resolve.bind(primitive(&resolveThisEnv));
  }
}

} // end namespace

} // end names
