
#include <string>

#include <iostream>

#include <ptr/Ptr.hpp>

using namespace sigma;

struct Foo {

  Foo(const std::string& s) : dummy(5), name(s) {
    std::cout << "Creating Foo(" << name << ")." << std::endl;
  }
  ~Foo() {
    std::cout << "Destroying Foo(" << name << ")." << std::endl;
  }

  /*
    "dummy" is to make sure memory layout isn't identical
    to "Bar".
  */
  int dummy; 
  std::string name;
};

struct Bar {

  Bar(const std::string& s) : name(s) {
    std::cout << "Creating Bar(" << name << ")." << std::endl;
  }
  ~Bar() {
    std::cout << "Destroying Bar(" << name << ")." << std::endl;
  }

  void print(std::ostream& out) {
    out << "Bar(" << name << ") ";
  }

  std::string name;
};

struct Baz : Foo, Bar {

  Baz(const std::string& s) : Foo(s), Bar(s) {
    std::cout << "Creating Baz(" << s << ")." << std::endl;
  }
  ~Baz() {
    std::cout << "Destroying Baz(Foo(" 
	      << Foo::name << "),Bar(" << Bar::name << "))." << std::endl;
  }
};


struct Circle {

  Circle(const std::string& n) : name(n) {
    std::cout << "Circle(" << name << ")" << std::endl;
  }
  ~Circle() {
    std::cout << "~Circle(" << name << ")" << std::endl;
  }

  Ptr<Circle> ptr;

  std::string name;
};

/*
  Test for collect of globals.
*/

Ptr<Bar> bar;

int main() {

  /* 
     Test for cast and subtype converter. Typically these 
     might be in headers in anonymous namespaces.
  */

  DeclareCastConverter<std::string,Baz> string_to_baz;
  DeclareSubtypeConverter<Baz,Bar> baz_to_bar;

  {
    //    Cell s(new string("first"));
    Ptr<Anything> s(new std::string("first"));
    bar = s; // build a Baz from string and cast Baz to bar
  }
  std::cout << "\nShould collect string cell." << std::endl;
  collect();

  if (!null(bar)) {
    std::cout << "\nGot a ";
    bar->print(std::cout);
    std::cout << "from a Cell(string).\n" << std::endl;
  } else {
    std::cout << "\nShould have gotten a Bar(first)." << std::endl;
  }


  /*
    Straight subtyping polymorphism doesn't use the type DAG.
  */

  Ptr<Baz> baz(new Baz("second"));
  bar = baz;
  if (!null(bar)) {
    std::cout << "\nGot a ";
    bar->print(std::cout);
    std::cout << "from a Ptr<Baz>" << std::endl;
  } else {
    std::cout << "Should have gotten a Bar(second).\n" << std::endl;
  }

  std::cout << "\nShould collect Baz(first)." << std::endl;
  collect();

  {
    Ptr<Circle> r = new Circle("root");
    {
      Ptr<Circle> a1 = new Circle("a1");
      Ptr<Circle> a2 = new Circle("a2");

      a1->ptr = a2;
      a2->ptr = a1;

      r->ptr = a1;
    }
    std::cout << "\nNothing to collect before Circle(root) out of scope." 
	      << std::endl;
    collect();
  }

  std::cout << "\nCollect all Circles after Circle(root) out of scope." 
	    << std::endl;
  collect();

  // causes a compiler error:
  // DeclareSubtypeConverter<Bar,Baz> bar_to_baz;

  // causes a compiler error:
  // Ptr<Foo> foo(new Bar("bar to foo"));

  // causes a compiler error:
  // Ptr<Foo> foo = bar;

  {
    Ptr<Foo> foo(new Baz("third"));
  }
  std::cout << "\nBaz(third) should be collected as a Baz \n"
	    << "even though destructor isn't virtual and \n"
	    << "even though assigned to a Ptr<Foo>." << std::endl;
    
  collect();

  {
    DeclareCastConverter<int,float> int_to_float;
    //    Cell i(new int(23));
    Ptr<Anything> i(new int(23));
    Ptr<float> f(i);
    if (!null(f)) {
      std::cout << "\nGot float " << (*f.get()) 
		<< " from Cell(int)." << std::endl;
    }
    try {
      float& k = deref<float>(i);
      std::cout << "k: " << k << std::endl;
    }
    catch(const BadType& error) {
      std::cout << "tried to dereference: " 
		<< error.expected.name() << std::endl
		<< "from a cell of type: " 
		<< error.found.name() << std::endl;
    }
#if 1
    try {
      DeclareCastConverter<float,int> float_to_int;
    }
    catch(const cell_detail::TypeDag::Circular& circular) {
      std::cout << "Converter from " << circular.from->name()
		<< " to " << circular.to->name() 
		<< " creates circularity." << std::endl;
    }
#endif
  }

  /*
    A Cell can be built from a Ptr<T>. Since Ptr<T> is derived from
    Cell, I don't know why it is needed, but there is an explicit
    constructor for this.
  */

  //  Cell cell = bar; // A cell can be built from a Ptr<T>
  //  Cell cell2 = cell; 

  Ptr<Anything> cell = bar; // A cell can be built from a Ptr<T>
  Ptr<Anything> cell2 = cell; 

  std::cout << "End of main. Final collect of baz." << std::endl;
}
