#include <symbol/Symbol.hpp>
#include <ptr/Type.hpp>
#include <ptr/Cell.hpp>
#include <env/Env.hpp>
#include <parse/Parser.hpp>
#include <generic/Primitive.hpp>
#include <generic/Generic.hpp>

#include <sigma/Bind.hpp>
#include <sigma/If.hpp>
#include <sigma/lex.hpp>
#include <sigma/parse.hpp>
#include <sigma/eval.hpp>
#include <sigma/resolve.hpp>
#include <sigma/env.hpp>
#include <sigma/arith.hpp>
#include <sigma/Cons.hpp>
#include <sigma/extend.hpp>
#include <sigma/Abort.hpp>
#ifdef SIGMA_EXPOSE_COLLECT
#include <sigma/collect.hpp>
#endif


#include <iostream>
#include <iterator>

#define REVISION "0.6"

using namespace sigma;

std::ostream& operator<<(std::ostream&, const Generic::NoMatch&);
std::ostream& operator<<(std::ostream&, const Method::NoMatch&);
std::ostream& operator<<(std::ostream&, const Types&);

int main() {

  Generic print = *getGeneric("print");

  std::cout << "Sigma " << REVISION << "\n? " << std::flush;

  Ptr<Context> context = new Context;

#if 0
  std::cin >> std::noskipws;
#else
  std::cin.unsetf(std::ios::skipws);
#endif

  typedef std::istream_iterator<char> InputIter;
  std::istream_iterator<char> in(std::cin);
  std::istream_iterator<char> end;

  typedef Dfa<char,ProdVal,Context>::BadEndOfInput BadEndOfInput;
  typedef Dfa<char,ProdVal,Context>::BadLetter BadLetter;
  typedef Parser<Context>::BadSymbol BadSymbol;
  typedef SigmaEnv::NotFound NotFound;

  Ptr<Dfa<char,ProdVal,Context> > dfa(makeDfa(buildRegExp()));
  Ptr<Lexer<char,ProdVal,Context,InputIter> > lexer = 
    new Lexer<char,ProdVal,Context,InputIter>(dfa,context,in,end);
  context->lexer = lexer;

  Parser<Context> parser(buildProductions());

  while (in!=end) {

    try {
     Ptr<Anything> exp = parser.parse(*lexer);
     Ptr<Anything> resolved_exp = resolve(theEnv(), exp);

#if 1
      Ptr<SigmaEnv> stack = make_sigma_env();
      Ptr<Anything> value = eval(stack, resolved_exp);
#else
      Ptr<Anything> value = eval(theEnv(), resolved_exp);
#endif

      Ptr<Bind> bind = value;
      if (!null(bind)) {
	theEnv() = extend(theEnv(), bind->formal.symbol, bind->exp);
	value = bind->exp;
      }

      if (!null(value)) {
	print(value);
	std::cout << " : " << value.type().name() 
		  << "\n? " << std::flush;
      } else {
	std::cout << "\n? " << std::flush;
      }
    }
    catch (const NotFound& error) {
      std::cout << "Undefined identifier: " << error.key
		<< "\n? " << std::flush;
    }
    catch(const BadEndOfInput&) {
      std::cout << "Unexpected end of input." << std::endl;
      return 0;
    }
    catch(const BadLetter& error) {
      std::cout << "Unexpected symbol: ";
      if (isprint(error.letter)) {
	std::cout << error.letter << std::endl;
      } else {
	std::cout << int(error.letter) << std::endl;
      }
      while ((in!=end) && (*in++!=';'));
      std::cout << "\n? " << std::flush;
      
    }
    catch (const BadSymbol& error) {
      if (error.symbol == Symbol("end_of_input")) {
	std::cout << "Bye." << std::endl;
	// Initialization issues require the environment to be
	// explicitly cleared if it's going to be cleaned up upon
	// exit. (Worth rethinking?)
	theEnv() = Ptr<SigmaEnv>();
	return 0;
      }
      std::cout << "Unexpected token: " << error.symbol << std::endl;
      if (!(error.symbol == Symbol(";"))) {
	  while ((in!=end) && (*in++!=';'));
      }
      std::cout << "\n? " << std::flush;
    }
    catch (const Generic::NoMatch& error) {
      std::cout << error;
      std::cout << "\n? " << std::flush;
    }
    catch (const Method::NoMatch& error) {
      std::cout << error;
      std::cout << "\n? " << std::flush;
    }
    catch (const BadApply& error) {
      std::cout << "Not a function: ";
      print(error.value);
      std::cout << " : " << error.value.type().name() << std::endl;
      std::cout << "\n? " << std::flush;
    }
    catch (const Type::NotFound& error) {
      std::cout << "Not a valid type name: " << error.name << std::endl;
      while ((in!=end) && (*in++!=';'));
      std::cout << "\n? " << std::flush;
    }
    catch (const If::NotBool& error) {
      std::cout << "Not a boolean: ";
      print(error.value);
      std::cout << "\n? " << std::flush;
    }
    catch (const Abort& abort) {
      std::cout << "Abort: ";
      print(abort.value);
      std::cout << "\n? " << std::flush;
    }
  }

  return 0;
}


std::ostream& operator<<(std::ostream& out, const Types& types) {

  out << "(";
  Types::const_iterator i = types.begin();
  if (i != types.end()) {
    while (true) {
      out << (*i).name();
      ++i;
      if (i == types.end()) {
	return out << ")";
      } else {
	out << ", ";
      }
    }
  }
  return out << ")";
}

std::ostream& operator<<(std::ostream& out, const Generic::NoMatch& error) {

  std::string fn = error.name.empty()?"anonymous function":error.name;

  out << "No match for " << fn << "\n"
      << "given argument type(s): "
      << error.types << "\n";

  out << "\nPossibilities are:\n";
  
  for (Forms::const_iterator j = error.forms.begin();
       j != error.forms.end();
       ++j) {

    const Form& form = *j;
    out << form.return_type.name() 
	<< " <- " 
	<< form.arg_types
	<< "\n";
  }
  return out << std::endl;
}


std::ostream& operator<<(std::ostream& out, const Method::NoMatch& error) {

  return out << "Method argument mismatch:\n"
	     << "Expected: " << error.expected << "\n"
	     << "Found: " << error.found << std::endl;

}
